"""Model for a review request draft."""

from __future__ import annotations

from typing import ClassVar, List, Optional, TYPE_CHECKING, Union

from django.contrib.auth.models import User
from django.db import models
from django.db.models import F
from django.utils import timezone
from django.utils.translation import gettext, gettext_lazy as _
from djblets.db.fields import ModificationTimestampField, RelationCounterField
from djblets.db.managers import ConcurrencyManager

from reviewboard.attachments.models import FileAttachment
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.diffviewer.models import DiffSet
from reviewboard.reviews.errors import NotModifiedError, PublishError
from reviewboard.reviews.fields import get_review_request_fields
from reviewboard.reviews.models.group import Group
from reviewboard.reviews.models.base_review_request_details import \
    BaseReviewRequestDetails
from reviewboard.reviews.models.review_request import ReviewRequest
from reviewboard.reviews.models.screenshot import Screenshot
from reviewboard.reviews.signals import review_request_published
from reviewboard.scmtools.errors import InvalidChangeNumberError

if TYPE_CHECKING:
    from datetime import datetime

    from django.contrib.auth.models import AbstractBaseUser, AnonymousUser
    from django.db.models.fields.related_descriptors import RelatedManager

    from reviewboard.attachments.models import FileAttachmentSequence
    from reviewboard.reviews.models.review_request import (
        FileAttachmentState,
        ReviewRequestFileAttachmentsData)
    from reviewboard.scmtools.core import ChangeSet


class ReviewRequestDraft(BaseReviewRequestDetails):
    """A draft of a review request.

    When a review request is being modified, a special draft copy of it is
    created containing all the details of the review request. This copy can
    be modified and eventually saved or discarded. When saved, the new
    details are copied back over to the originating ReviewRequest.
    """

    summary = models.CharField(
        _('summary'),
        max_length=BaseReviewRequestDetails.MAX_SUMMARY_LENGTH)

    owner = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        verbose_name=_('owner'),
        null=True,
        related_name='draft')
    review_request = models.ForeignKey(
        ReviewRequest,
        on_delete=models.CASCADE,
        related_name='draft',
        verbose_name=_('review request'),
        unique=True)
    last_updated = ModificationTimestampField(
        _('last updated'))
    diffset = models.ForeignKey(
        DiffSet,
        on_delete=models.SET_NULL,
        verbose_name=_('diff set'),
        blank=True,
        null=True,
        related_name='review_request_draft')
    changedesc = models.ForeignKey(
        ChangeDescription,
        on_delete=models.SET_NULL,
        verbose_name=_('change description'),
        blank=True,
        null=True,
        related_name='review_request_draft')
    target_groups = models.ManyToManyField(
        Group,
        related_name='drafts',
        verbose_name=_('target groups'),
        blank=True)
    target_people = models.ManyToManyField(
        User,
        verbose_name=_('target people'),
        related_name='directed_drafts',
        blank=True)
    screenshots = models.ManyToManyField(
        Screenshot,
        related_name='drafts',
        verbose_name=_('screenshots'),
        blank=True)
    inactive_screenshots = models.ManyToManyField(
        Screenshot,
        verbose_name=_('inactive screenshots'),
        related_name='inactive_drafts',
        blank=True)

    file_attachments = models.ManyToManyField(
        FileAttachment,
        related_name='drafts',
        verbose_name=_('file attachments'),
        blank=True)
    inactive_file_attachments = models.ManyToManyField(
        FileAttachment,
        verbose_name=_('inactive files'),
        related_name='inactive_drafts',
        blank=True)

    submitter = property(lambda self: self.owner or
                         self.review_request.owner)
    repository = property(lambda self: self.review_request.repository)
    local_site = property(lambda self: self.review_request.local_site)

    depends_on = models.ManyToManyField('ReviewRequest',
                                        blank=True,
                                        verbose_name=_('Dependencies'),
                                        related_name='draft_blocks')

    screenshots_count = RelationCounterField(
        'screenshots',
        verbose_name=_('screenshots count'))

    inactive_screenshots_count = RelationCounterField(
        'inactive_screenshots',
        verbose_name=_('inactive screenshots count'))

    file_attachments_count = RelationCounterField(
        'file_attachments',
        verbose_name=_('file attachments count'))

    inactive_file_attachments_count = RelationCounterField(
        'inactive_file_attachments',
        verbose_name=_('inactive file attachments count'))

    # Set this up with a ConcurrencyManager to help prevent race conditions.
    objects: ClassVar[ConcurrencyManager] = ConcurrencyManager()

    commit = property(lambda self: self.commit_id,
                      lambda self, value: setattr(self, 'commit_id', value))

    def get_latest_diffset(self) -> Optional[DiffSet]:
        """Return the diffset for this draft.

        Returns:
            reviewboard.diffviewer.models.DiffSet:
            The diffset for the draft.
        """
        return self.diffset

    def is_accessible_by(
        self,
        user: Union[AbstractBaseUser, AnonymousUser],
    ) -> bool:
        """Return whether or not the user can access this draft.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            Whether the user can access the draft.
        """
        return self.is_mutable_by(user)

    def is_mutable_by(
        self,
        user: Union[AbstractBaseUser, AnonymousUser],
    ) -> bool:
        """Return whether or not the user can modify this draft.

        Args:
            user (django.contrib.auth.models.User):
                The user to check.

        Returns:
            bool:
            Whether the user can modify the draft.
        """
        return self.review_request.is_mutable_by(user)

    def get_file_attachments_data(
        self,
        *,
        active_attachments: Optional[FileAttachmentSequence] = None,
        inactive_attachments: Optional[FileAttachmentSequence] = None,
        draft_active_attachments: Optional[FileAttachmentSequence] = None,
        draft_inactive_attachments: Optional[FileAttachmentSequence] = None,
        **kwargs,
    ) -> ReviewRequestFileAttachmentsData:
        """Return data about a review request and its draft's file attachments.

        This returns sets of the active and inactive file attachment IDs
        that are attached to the review request or its draft. This data is
        used in :py:meth:`ReviewRequest.get_file_attachment_state()
        <reviewboard.reviews.models.review_request.ReviewRequest
        .get_file_attachment_state>`.

        The active and inactive file attachments on the review request and its
        draft may be passed in to avoid fetching them again if they've already
        been fetched elsewhere.

        The returned data will be cached for future lookups.

        Version Added:
            6.0

        Args:
            active_attachments (list, optional):
                The list of active file attachments on the review request.

            inactive_attachments (list, optional):
                The list of inactive file attachments on the review request.

            draft_active_attachments (list, optional):
                The list of active file attachments on the review request
                draft.

            draft_inactive_attachments (list, optional):
                The list of inactive file attachments on the review request
                draft.

            **kwargs (dict, unused):
                Keyword arguments for future expansion.

        Returns:
            ReviewRequestFileAttachmentsData:
                The data about the file attachments on a review request and
                its draft.
        """
        return self.get_review_request().get_file_attachments_data(
            active_attachments=active_attachments,
            inactive_attachments=inactive_attachments,
            draft_active_attachments=draft_active_attachments,
            draft_inactive_attachments=draft_inactive_attachments)

    def get_file_attachment_state(
        self,
        file_attachment: FileAttachment,
    ) -> FileAttachmentState:
        """Get the state of a file attachment attached to this review request.

        Version Added:
            6.0

        Args:
            file_attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment whose state will be returned.

        Returns:
            FileAttachmentState:
            The file attachment state.
        """
        return self.get_review_request().get_file_attachment_state(
            file_attachment)

    @staticmethod
    def create(review_request, changedesc=None):
        """Create a draft based on a review request.

        This will copy over all the details of the review request that
        we care about. If a draft already exists for the review request,
        the draft will be returned.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to fetch or create the draft from.

            changedesc (reviewboard.changedescs.models.ChangeDescription):
                A custom change description to set on the draft. This will
                always be set, overriding any previous one if already set.

        Returns:
            ReviewRequestDraft:
            The resulting draft.
        """
        draft, draft_is_new = \
            ReviewRequestDraft.objects.get_or_create(
                review_request=review_request,
                defaults={
                    'changedesc': changedesc,
                    'extra_data': review_request.extra_data or {},
                    'summary': review_request.summary,
                    'description': review_request.description,
                    'testing_done': review_request.testing_done,
                    'bugs_closed': review_request.bugs_closed,
                    'branch': review_request.branch,
                    'description_rich_text':
                        review_request.description_rich_text,
                    'testing_done_rich_text':
                        review_request.testing_done_rich_text,
                    'rich_text': review_request.rich_text,
                    'commit_id': review_request.commit_id,
                })

        if (changedesc is None and
            draft.changedesc_id is None and
            review_request.public):
            changedesc = ChangeDescription.objects.create()

        if changedesc is not None and draft.changedesc_id != changedesc.pk:
            old_changedesc_id = draft.changedesc_id
            draft.changedesc = changedesc
            draft.save(update_fields=('changedesc',))

            if old_changedesc_id is not None:
                ChangeDescription.objects.filter(pk=old_changedesc_id).delete()

        if draft_is_new:
            rels_to_update = [
                ('depends_on', 'to_reviewrequest_id', 'from_reviewrequest_id'),
                ('target_groups', 'group_id', 'reviewrequest_id'),
                ('target_people', 'user_id', 'reviewrequest_id'),
            ]

            if review_request.screenshots_count > 0:
                review_request.screenshots.update(draft_caption=F('caption'))
                rels_to_update.append(('screenshots', 'screenshot_id',
                                       'reviewrequest_id'))

            if review_request.inactive_screenshots_count > 0:
                review_request.inactive_screenshots.update(
                    draft_caption=F('caption'))
                rels_to_update.append(('inactive_screenshots', 'screenshot_id',
                                       'reviewrequest_id'))

            if review_request.file_attachments_count > 0:
                review_request.file_attachments.update(
                    draft_caption=F('caption'))
                rels_to_update.append(('file_attachments', 'fileattachment_id',
                                       'reviewrequest_id'))

            if review_request.inactive_file_attachments_count > 0:
                review_request.inactive_file_attachments.update(
                    draft_caption=F('caption'))
                rels_to_update.append(('inactive_file_attachments',
                                       'fileattachment_id',
                                       'reviewrequest_id'))

            for rel_field, id_field, lookup_field, in rels_to_update:
                # We don't need to query the entirety of each object, and
                # we'd like to avoid any JOINs. So, we'll be using the
                # M2M 'through' tables to perform lookups of the related
                # models' IDs.
                items = list(
                    getattr(review_request, rel_field).through.objects
                    .filter(**{lookup_field: review_request.pk})
                    .values_list(id_field, flat=True)
                )

                if items:
                    # Note that we're using add() instead of directly
                    # assigning the value. This lets us avoid a query that
                    # Django would perform to determine if it needed to clear
                    # out any existing values. Since we know this draft is
                    # new, there's no point in doing that.
                    getattr(draft, rel_field).add(*items)

        review_request._draft = draft

        return draft

    def publish(
        self,
        review_request: Optional[ReviewRequest] = None,
        user: Optional[User] = None,
        trivial: bool = False,
        send_notification: bool = True,
        validate_fields: bool = True,
        timestamp: Optional[datetime] = None,
    ) -> ChangeDescription:
        """Publish this draft.

        This is an internal method. Programmatic publishes should use
        :py:meth:`reviewboard.reviews.models.review_request.ReviewRequest.publish`
        instead.

        This updates and returns the draft's ChangeDescription, which
        contains the changed fields. This is used by the e-mail template
        to tell people what's new and interesting.

        The keys that may be saved in ``fields_changed`` in the
        ChangeDescription are:

        *  ``submitter``
        *  ``summary``
        *  ``description``
        *  ``testing_done``
        *  ``bugs_closed``
        *  ``depends_on``
        *  ``branch``
        *  ``target_groups``
        *  ``target_people``
        *  ``screenshots``
        *  ``screenshot_captions``
        *  ``diff``
        *  Any custom field IDs

        Each field in 'fields_changed' represents a changed field. This will
        save fields in the standard formats as defined by the
        'ChangeDescription' documentation, with the exception of the
        'screenshot_captions' and 'diff' fields.

        For the 'screenshot_captions' field, the value will be a dictionary
        of screenshot ID/dict pairs with the following fields:

        * ``old``: The old value of the field
        * ``new``: The new value of the field

        For the ``diff`` field, there is only ever an ``added`` field,
        containing the ID of the new diffset.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest, optional):
                The review request associated with this diff. If not provided,
                it will be looked up.

            user (django.contrib.auth.models.User, optional):
                The user publishing the draft. If not provided, this defaults
                to the review request submitter.

            trivial (bool, optional):
                Whether or not this is a trivial publish.

                Trivial publishes do not result in e-mail notifications.

            send_notification (bool, optional):
                Whether or not this will emit the
                :py:data:`reviewboard.reviews.signals.review_request_published`
                signal.

                This parameter is intended for internal use **only**.

            validate_fields (bool, optional):
                Whether or not the fields should be validated.

                This should only be ``False`` in the case of programmatic
                publishes, e.g., from close as submitted hooks.

            timestamp (datetime.datetime, optional):
                The datetime that should be used for all timestamps for objects
                published
                (:py:class:`~reviewboard.diffviewer.models.diff_set.DiffSet`,
                :py:class:`~reviewboard.changedescs.models.ChangeDescription`)
                over the course of the method.

        Returns:
            reviewboard.changedescs.models.ChangeDescription:
            The change description that results from this publish (if any).

            If this is an initial publish, there will be no change description
            (and this function will return ``None``).
        """
        if timestamp is None:
            timestamp = timezone.now()

        if not review_request:
            review_request = self.review_request

        assert review_request is not None

        if not self.changedesc and review_request.public:
            self.changedesc = ChangeDescription()

        changedesc = self.changedesc

        if not user:
            if changedesc:
                user = changedesc.get_user(self)
            else:
                user = review_request.submitter

        self.copy_fields_to_request(review_request)

        diffset = self.diffset

        # If no changes were made, raise exception and do not save
        if changedesc and not changedesc.has_modified_fields():
            raise NotModifiedError()

        if validate_fields:
            if not (self.target_groups.exists() or
                    self.target_people.exists()):
                raise PublishError(
                    gettext('There must be at least one reviewer before this '
                            'review request can be published.'))

            if not review_request.summary.strip():
                raise PublishError(
                    gettext('The draft must have a summary.'))

            if not review_request.description.strip():
                raise PublishError(
                    gettext('The draft must have a description.'))

            if (review_request.created_with_history and
                diffset and
                diffset.commit_count == 0):
                raise PublishError(
                    gettext('There are no commits attached to the diff.'))

        if diffset:
            if (review_request.created_with_history and not
                diffset.is_commit_series_finalized):
                raise PublishError(gettext(
                    'This commit series is not finalized.'))

            diffset.history = review_request.diffset_history
            diffset.timestamp = timestamp
            diffset.save(update_fields=('history', 'timestamp'))

        if changedesc:
            changedesc.user = user
            changedesc.timestamp = timestamp
            changedesc.public = True
            changedesc.save()
            review_request.changedescs.add(changedesc)

        review_request.description_rich_text = self.description_rich_text
        review_request.testing_done_rich_text = self.testing_done_rich_text
        review_request.rich_text = self.rich_text
        review_request.save()

        if send_notification:
            review_request_published.send(sender=type(review_request),
                                          user=user,
                                          review_request=review_request,
                                          trivial=trivial,
                                          changedesc=changedesc)

        return self.changedesc

    def update_from_commit_id(self, commit_id):
        """Update the data from a server-side changeset.

        If the commit ID refers to a pending changeset on an SCM which stores
        such things server-side (like Perforce), the details like the summary
        and description will be updated with the latest information.

        If the change number is the commit ID of a change which exists on the
        server, the summary and description will be set from the commit's
        message, and the diff will be fetched from the SCM.

        Args:
            commit_id (unicode):
                The commit ID or changeset ID that the draft will update
                from.

        Returns:
            list of unicode:
            The list of draft fields that have been updated from the commit.

        Raises:
            reviewboard.hostingsvcs.errors.HostingServiceError:
                The hosting service backing the repository encountered an
                error.

            reviewboard.scmtools.errors.InvalidChangeNumberError:
                The commit ID could not be found in the repository.

            reviewboard.scmtools.errors.SCMError:
                The repository tool encountered an error.

            NotImplementedError:
                The repository does not support fetching information from
                commit IDs.
        """
        changeset = None

        if self.repository.supports_pending_changesets:
            scmtool = self.repository.get_scmtool()
            changeset = scmtool.get_changeset(commit_id, allow_empty=True)

        if changeset and changeset.pending:
            return self.update_from_pending_change(commit_id, changeset)
        elif self.repository.supports_post_commit:
            return self.update_from_committed_change(commit_id)
        else:
            if changeset:
                raise InvalidChangeNumberError()
            else:
                raise NotImplementedError()

    def update_from_pending_change(
        self,
        commit_id: str,
        changeset: ChangeSet,
    ) -> List[str]:
        """Update the data from a server-side pending changeset.

        This will fetch the metadata from the server and update the fields on
        the draft.

        Version Changed:
            6.0:
            Added support for setting ``extra_data``.

        Args:
            commit_id (str):
                The changeset ID that the draft will update from.

            changeset (reviewboard.scmtools.core.ChangeSet):
                The changeset information to update from.

        Returns:
            list of str:
            The list of draft fields that have been updated from the change.

        Raises:
            reviewboard.scmtools.errors.InvalidChangeNumberError:
                A changeset could not be found.
        """
        if not changeset:
            raise InvalidChangeNumberError()

        # If the SCM supports changesets, they should always include a number,
        # summary and description, parsed from the changeset description. Some
        # specialized systems may support the other fields, but we don't want
        # to clobber the user-entered values if they don't.
        self.commit = commit_id
        description = changeset.description
        testing_done = changeset.testing_done

        self.summary = changeset.summary
        self.description = description
        self.description_rich_text = False

        modified_fields = [
            'commit_id', 'summary', 'description', 'description_rich_text',
        ]

        if testing_done:
            self.testing_done = testing_done
            self.testing_done_rich_text = False
            modified_fields += ['testing_done', 'testing_done_rich_text']

        if changeset.branch:
            self.branch = changeset.branch
            modified_fields.append('branch')

        if changeset.bugs_closed:
            self.bugs_closed = ','.join(changeset.bugs_closed)
            modified_fields.append('bugs_closed')

        if changeset.extra_data:
            if self.extra_data is None:
                self.extra_data = {}

            self.extra_data.update(changeset.extra_data)
            modified_fields.append('extra_data')

        return modified_fields

    def update_from_committed_change(self, commit_id):
        """Update from a committed change present on the server.

        Fetches the commit message and diff from the repository and sets the
        relevant fields.

        Args:
            commit_id (unicode):
                The commit ID to update from.

        Returns:
            list of unicode:
            The list of draft fields that have been updated from the commit
            message.
        """
        commit = self.repository.get_change(commit_id)
        summary, message = commit.split_message()
        message = message.strip()

        self.commit = commit_id
        self.summary = summary.strip()

        self.description = message
        self.description_rich_text = False

        self.diffset = DiffSet.objects.create_from_data(
            repository=self.repository,
            diff_file_name='diff',
            diff_file_contents=commit.diff,
            parent_diff_file_name=None,
            parent_diff_file_contents=None,
            diffset_history=None,
            basedir='/',
            request=None,
            base_commit_id=commit.parent,
            check_existence=False)

        # Compute a suitable revision for the diffset.
        self.diffset.update_revision_from_history(
            self.review_request.diffset_history)
        self.diffset.save(update_fields=('revision',))

        return [
            'commit_id',
            'description',
            'description_rich_text',
            'diffset',
            'summary',
        ]

    def copy_fields_to_request(
        self,
        review_request: ReviewRequest,
    ) -> None:
        """Copy draft fields to the review request.

        This will loop through all fields on the review request, copying any
        changes from the draft to the review request, in preparation for a
        publish.

        Args:
            review_request (reviewboard.reviews.models.review_request.
                            ReviewRequest):
                The review request to copy the fields to.
        """
        changedesc = self.changedesc

        for field_cls in get_review_request_fields():
            field = field_cls(review_request)

            if field.can_record_change_entry:
                old_value = field.load_value(review_request)
                new_value = field.load_value(self)

                if field.has_value_changed(old_value, new_value):
                    field.propagate_data(self)

                    if changedesc:
                        field.record_change_entry(changedesc,
                                                  old_value, new_value)

        # Screenshots and file attachments are a bit special. The list of
        # associated items can change, but so can captions within each.
        #
        # Fortunately, both of these types of models have near-identical
        # signatures, so we can reuse the same code for both.
        self._copy_attachments_to_review_request(
            changedesc=changedesc,
            changedesc_captions_field='screenshot_captions',
            changedesc_items_field='screenshots',
            changedesc_item_name_field='caption',
            review_request_models_active=review_request.screenshots,
            review_request_models_active_count=(
                review_request.screenshots_count),
            review_request_models_inactive=(
                review_request.inactive_screenshots),
            review_request_models_inactive_count=(
                review_request.inactive_screenshots_count),
            draft_models_active=self.screenshots,
            draft_models_active_count=self.screenshots_count,
            draft_models_inactive=self.inactive_screenshots,
            draft_models_inactive_count=self.inactive_screenshots_count)

        self._copy_attachments_to_review_request(
            changedesc=changedesc,
            changedesc_captions_field='file_captions',
            changedesc_items_field='files',
            changedesc_item_name_field='display_name',
            review_request_models_active=review_request.file_attachments,
            review_request_models_active_count=(
                review_request.file_attachments_count),
            review_request_models_inactive=(
                review_request.inactive_file_attachments),
            review_request_models_inactive_count=(
                review_request.inactive_file_attachments_count),
            draft_models_active=self.file_attachments,
            draft_models_active_count=self.file_attachments_count,
            draft_models_inactive=self.inactive_file_attachments,
            draft_models_inactive_count=self.inactive_file_attachments_count)

    def _copy_attachments_to_review_request(
        self,
        *,
        changedesc: Optional[ChangeDescription],
        changedesc_captions_field: str,
        changedesc_items_field: str,
        changedesc_item_name_field: str,
        review_request_models_active: RelatedManager,
        review_request_models_active_count: int,
        review_request_models_inactive: RelatedManager,
        review_request_models_inactive_count: int,
        draft_models_active: RelatedManager,
        draft_models_active_count: int,
        draft_models_inactive: RelatedManager,
        draft_models_inactive_count: int,
    ) -> None:
        """Copy over screenshots or file attachments to a review request.

        This takes care to copy over any screenshots/file attachments (as
        provided by the caller) from this draft to a review request. In the
        process, any changes to captions will be recorded in the change
        description.

        The logic attempts to minimize the number of database queries
        necessary to perform these updates.

        Args:
            changedesc (reviewboard.changedescs.models.ChangeDescription):
                The change description to update.

            changedesc_items_field (str):
                The field to set in the change description for any item
                changes.

            changedesc_captions_field (str):
                The field to set in the change description for any caption
                changes.

            changedesc_item_name_field (str):
                The name of the field representing the name of an item.

            review_request_models_active (django.db.models.manager.
                                          RelatedManager):
                The manager for the relation managing active items on the
                draft.

            review_request_models_active_count (int):
                The number of active items on the draft.

            review_request_models_inactive (django.db.models.manager.
                                            RelatedManager):
                The manager for the relation managing inactive items on the
                draft.

            review_request_models_inactive_count (int):
                The number of inactive items on the draft.

            draft_models_active (django.db.models.manager. RelatedManager):
                The manager for the relation managing active items on the
                draft.

            draft_models_active_count (int):
                The number of active items on the draft.

            draft_models_inactive (django.db.models.manager. RelatedManager):
                The manager for the relation managing inactive items on the
                draft.

            draft_models_inactive_count (int):
                The number of inactive items on the draft.
        """
        if (review_request_models_active_count > 0 or
            draft_models_active_count > 0):
            old_ids = set(
                review_request_models_active.values_list('pk', flat=True)
            )
            new_items = list(
                draft_models_active
                .only('pk', 'caption', 'draft_caption')
            )
            caption_changes = {}

            # Update the captions for each item. We won't be considering any
            # that were removed from the review request.
            for item in new_items:
                if item.caption != item.draft_caption:
                    if (item.pk in old_ids and
                        not (isinstance(item, FileAttachment) and
                             item.is_from_diff)):
                        caption_changes[item.pk] = {
                            'old': (item.caption,),
                            'new': (item.draft_caption,),
                        }

                    item.caption = item.draft_caption
                    item.save(update_fields=['caption'])

            if caption_changes and changedesc:
                changedesc.fields_changed[changedesc_captions_field] = \
                    caption_changes

            # If the list of IDs have changed, record the old/new items in the
            # change description, and then reset the list on the review request
            # to match the draft.
            new_ids = {
                item.pk
                for item in new_items
            }

            if new_ids.symmetric_difference(old_ids):
                if changedesc:
                    changedesc.record_field_change(
                        field=changedesc_items_field,
                        old_value=review_request_models_active.all(),
                        new_value=draft_models_active.all(),
                        name_field=changedesc_item_name_field)

                # Note that we explicitly want to clear this to preserve the
                # insertion order.
                review_request_models_active.set(draft_models_active.all(),
                                                 clear=True)

        if (review_request_models_inactive_count > 0 or
            draft_models_inactive_count > 0):
            # There's no change description entry required for this field.
            # We can just copy them.
            review_request_models_inactive.set(draft_models_inactive.all())

    def get_review_request(self):
        """Returns the associated review request."""
        return self.review_request

    class Meta:
        app_label = 'reviews'
        db_table = 'reviews_reviewrequestdraft'
        ordering = ['-last_updated']
        verbose_name = _('Review Request Draft')
        verbose_name_plural = _('Review Request Drafts')



"""URLs for processing license-related state.

Version Added:
    8.0
"""

from __future__ import annotations

from django.urls import path

from reviewboard.licensing.views import LicensesView


urlpatterns = [
    path('',
         LicensesView.as_view(),
         name='admin-licenses'),
]



import datetime
import time

from django.core.cache import cache
from django.contrib.auth.models import User
from django.db.models.aggregates import Count
from django.db.models.signals import post_save, post_delete
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import gettext_lazy as _
from djblets.cache.backend import cache_memoize
from djblets.registries.registry import (ALREADY_REGISTERED,
                                         ATTRIBUTE_REGISTERED,
                                         DEFAULT_ERRORS,
                                         NOT_REGISTERED,
                                         OrderedRegistry,
                                         UNREGISTER)
from djblets.util.decorators import augment_method_from

from reviewboard import get_manual_url
from reviewboard.admin.cache_stats import get_cache_stats
from reviewboard.changedescs.models import ChangeDescription
from reviewboard.reviews.models import Comment, Group, Review, ReviewRequest
from reviewboard.scmtools.models import Repository


class BaseAdminWidget(object):
    """The base class for an Administration Dashboard widget.

    Widgets appear in the Administration Dashboard and can display useful
    information on the system, links to other pages, or even fetch data
    from external sites.

    There are a number of built-in widgets, but extensions can provide their
    own.

    Version Added::
        4.0:
        Introduced a a replacement for the legacy :py:attr:`Widget` class.
    """

    #: The unique ID of the widget.
    widget_id = None

    #: The name of the widget.
    #:
    #: This will be shown at the top of the widget.
    name = None

    #: The name of the template used to render the widget.
    template_name = 'admin/admin_widget.html'

    #: Additional CSS classes to apply to the widget.
    #:
    #: If set, this must be a string with a space-separated list of CSS
    #: classes.
    css_classes = None

    #: The name of the JavaScript view rendering the widget.
    js_view_class = 'RB.Admin.WidgetView'

    #: The name of the JavaScript model handling widget state.
    js_model_class = 'RB.Admin.Widget'

    def __init__(self):
        """Initialize the widget."""
        self.dom_id = None

    def can_render(self, request):
        """Return whether the widget can be rendered in the dashboard.

        Subclasses can override this to make certain widgets conditional.
        By default, widgets can always be rendered.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            bool:
            ``True``, always.
        """
        return True

    def get_js_model_attrs(self, request):
        """Return attributes to pass to the JavaScript model.

        These attributes will be passed to the widget model when instantiated.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        return {}

    def get_js_model_options(self, request):
        """Return options to pass to the JavaScript model.

        These options will be passed to the widget model when instantiated.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The options to pass to the model.
        """
        return {}

    def get_js_view_options(self, request):
        """Return options to pass to the JavaScript view.

        These options will be passed to the widget view when instantiated.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The options to pass to the view.
        """
        return {}

    def get_extra_context(self, request):
        """Return extra context for the template.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Extra context to pass to the template.
        """
        return {}

    def render(self, request):
        """Render the widget to a string.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.utils.safestring.SafeText:
            The rendered widget HTML.
        """
        return render_to_string(
            template_name=self.template_name,
            context=dict({
                'widget': self,
            }, **self.get_extra_context(request)),
            request=request)


class AdminWidgetsRegistry(OrderedRegistry):
    """The registry managing all administration dashboard widgets."""

    lookup_attrs = ('widget_id',)

    default_errors = dict(DEFAULT_ERRORS, **{
        ALREADY_REGISTERED: _(
            'Could not register the administration widget %(item)s. This '
            'widget is already registered or its ID conflicts with another '
            'widget.'
        ),
        ATTRIBUTE_REGISTERED: _(
            'Could not register the administration widget %(item)s: Another '
            'widget (%(duplicate)s) is already registered with the same ID.'
        ),
        NOT_REGISTERED: _(
            'No administration widget was found with an ID of '
            '"%(attr_value)s".'
        ),
        UNREGISTER: _(
            'Could not unregister the administration widget %(item)s: This '
            'widget has not been registered.'
        ),
    })

    @augment_method_from(OrderedRegistry)
    def register(self, admin_widget_cls):
        """Register a new administration widget class.

        Args:
            admin_widget_cls (type):
                The widget class to register. This must be a subclass of
                :py:class:`BaseAdminWidget`.

        Raises:
            djblets.registries.errors.RegistrationError:
                The :py:attr:`BaseAdminWidget.widget_id` value is missing on
                the class.

            djblets.registries.errors.AlreadyRegisteredError:
                This widget, or another with the same ID, was already
                registered.
        """
        pass

    @augment_method_from(OrderedRegistry)
    def unregister(self, admin_widget_cls):
        """Unregister an administration widget class.

        Args:
            admin_widget_cls (type):
                The widget class to unregister. This must be a subclass of
                :py:class:`BaseAdminWidget`.

        Raises:
            djblets.registries.errors.ItemLookupError:
                This widget was not registered.
        """
        pass

    def get_widget(self, widget_id):
        """Return a widget class with the specified ID.

        Args:
            widget_id (unicode):
                The ID of the widget to return.

        Returns:
            type:
            The subclass of :py:class:`BaseAdminWidget` that was registered
            with the given ID, if found. If the widget was not found, this
            will return ``None``.
        """
        try:
            return self.get('widget_id', widget_id)
        except self.lookup_error_class:
            return None

    def get_defaults(self):
        """Return the default widgets for the administration dashboard.

        Returns:
            list of type:
            The list of default widgets.
        """
        return [
            NewsWidget,
            ActivityGraphWidget,
            RepositoriesWidget,
            UserActivityWidget,
            ServerCacheWidget,
        ]


def get_sync_num():
    """Get the sync_num, which is number to sync.

    sync_num is number of update and initialized to 1 every day.
    """
    KEY = datetime.date.today()
    cache.add(KEY, 1)
    return cache.get(KEY)


def _increment_sync_num(*args, **kwargs):
    """Increment the sync_num."""
    KEY = datetime.date.today()

    if cache.get(KEY) is not None:
        cache.incr(KEY)


class UserActivityWidget(BaseAdminWidget):
    """A widget displaying stats on how often users interact with Review Board.

    This is displayed as a pie graph, with a legend alongside it breaking
    down the activity into 1-6 day, 7-29 day, 30-59 day, 60-89 day, and 90+
    day ranges.
    """

    widget_id = 'user-activity-widget'
    name = _('User Activity')
    js_view_class = 'RB.Admin.UserActivityWidgetView'
    css_classes = 'rb-c-admin-user-activity-widget'

    def get_js_model_attrs(self, request):
        """Return data for the JavaScript model.

        This will calculate the user activity in the various time ranges,
        and return the data for use in a rendered chart.

        Args:
            request (django.http.HttpRequest, unused):
                The HTTP request from the client.

        Returns:
            dict:
            Data for the JavaScript model,.
        """
        now = timezone.now()
        users = User.objects

        week = datetime.timedelta(days=7)
        day = datetime.timedelta(days=1)
        month = datetime.timedelta(days=30)
        two_months = datetime.timedelta(days=60)
        three_months = datetime.timedelta(days=90)

        one_day = (now - week, now + day)
        seven_days = (now - month, now - week)
        thirty_days = (now - two_months, now - month)
        sixty_days = (now - three_months, now - two_months)
        ninety_days = now - three_months

        return {
            'now': users.filter(last_login__range=one_day).count(),
            'sevenDays': users.filter(last_login__range=seven_days).count(),
            'thirtyDays': users.filter(last_login__range=thirty_days).count(),
            'sixtyDays': users.filter(last_login__range=sixty_days).count(),
            'ninetyDays': users.filter(last_login__lte=ninety_days).count(),
            'total': users.count()
        }


class RepositoriesWidget(BaseAdminWidget):
    """A widget displaying the most recent repositories.

    This widget displays a grid of the most recent repositories and their
    services/types.
    """

    #: The maximum number of repositories shown in the widget.
    MAX_REPOSITORIES = 8

    widget_id = 'repositories-widget'
    name = _('Repositories')
    css_classes = 'rb-c-admin-repositories-widget'
    template_name = 'admin/widgets/repositories.html'

    def get_extra_context(self, request):
        """Return extra context for the template.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Extra context to pass to the template.
        """
        extra_context = cache_memoize(
            'admin-widget-repos-data',
            lambda: self._get_repositories_data(request))
        extra_context['add_repo_docs_url'] = \
            '%sadmin/configuration/repositories/' % get_manual_url()

        return extra_context

    def _get_repositories_data(self, request):
        """Return data on the repositories for the widget.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Information on the repositories, for use as template context.
        """
        queryset = Repository.objects.accessible(request.user)
        total_repositories = queryset.count()

        if total_repositories > 0:
            queryset = (
                queryset
                .only('id', 'hosting_account_id', 'name', 'tool_id')
                .select_related('hosting_account')
                [:self.MAX_REPOSITORIES]
            )
        else:
            queryset = Repository.objects.none()

        repositories_info = []

        for repository in queryset:
            hosting_service = repository.hosting_service
            scmtool_class = repository.scmtool_class
            service_name = scmtool_class.name

            if hosting_service:
                service_name = _('%(hosting_service)s (%(tool)s)') % {
                    'hosting_service': hosting_service.name,
                    'tool': service_name,
                }

            repositories_info.append({
                'id': repository.pk,
                'name': repository.name,
                'service': service_name,
            })

        return {
            'repositories': repositories_info,
            'total_repositories': total_repositories,
        }


class ServerCacheWidget(BaseAdminWidget):
    """Cache statistics widget.

    Displays a list of memcached statistics, if available.
    """

    widget_id = 'server-cache-widget'
    name = _('Server Cache')
    template_name = 'admin/widgets/server_cache.html'

    def get_extra_context(self, request):
        """Return extra context for the template.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            Extra context to pass to the template.
        """
        context = super(ServerCacheWidget, self).get_extra_context(request)

        uptime = {}
        cache_stats = get_cache_stats()

        if cache_stats:
            for hosts, stats in cache_stats:
                uptime_secs = stats['uptime']

                if uptime_secs > 86400:
                    uptime['value'] = uptime_secs / 60 / 60 / 24
                    uptime['unit'] = _('days')
                elif uptime_secs > 3600:
                    uptime['value'] = uptime_secs / 60 / 60
                    uptime['unit'] = _('hours')
                else:
                    uptime['value'] = uptime_secs / 60
                    uptime['unit'] = _('minutes')

        context['cache_stats'] = cache_stats
        context['uptime'] = uptime

        return context


class NewsWidget(BaseAdminWidget):
    """A widget displaying the latest Review Board news headlines."""

    widget_id = 'news'
    name = _('Review Board News')
    css_classes = 'rb-c-admin-news-widget'
    js_view_class = 'RB.Admin.NewsWidgetView'
    js_model_class = 'RB.Admin.NewsWidget'

    def get_js_model_attrs(self, request):
        """Return attributes to pass to the JavaScript model.

        These contain URLs for the RSS feed and the public news page.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        return {
            'rssURL': 'https://www.reviewboard.org/news/feed/rss/latest/',
            'newsURL': 'https://www.reviewboard.org/news/',
            'subscribeURL': 'https://www.reviewboard.org/mailing-lists/',
        }


def dynamic_activity_data(request):
    """Large database activity widget helper.

    This method serves as a helper for the activity widget, it's used with for
    AJAX requests based on date ranges passed to it.
    """
    direction = request.GET.get('direction')
    range_end = request.GET.get('range_end')
    range_start = request.GET.get('range_start')
    days_total = 30

    # Convert the date from the request.
    #
    # This takes the date from the request in YYYY-MM-DD format and
    # converts into a format suitable for QuerySet later on.
    if range_end:
        range_end = datetime.datetime.fromtimestamp(
            time.mktime(time.strptime(range_end, "%Y-%m-%d")))

    if range_start:
        range_start = datetime.datetime.fromtimestamp(
            time.mktime(time.strptime(range_start, "%Y-%m-%d")))

    if direction == "next" and range_end:
        new_range_start = range_end
        new_range_end = \
            new_range_start + datetime.timedelta(days=days_total)
    elif direction == "prev" and range_start:
        new_range_start = range_start - datetime.timedelta(days=days_total)
        new_range_end = range_start
    elif direction == "same" and range_start and range_end:
        new_range_start = range_start
        new_range_end = range_end
    else:
        new_range_end = datetime.datetime.now() + datetime.timedelta(days=1)
        new_range_start = new_range_end - datetime.timedelta(days=days_total)

    current_tz = timezone.get_current_timezone()
    new_range_start = timezone.make_aware(new_range_start, current_tz)
    new_range_end = timezone.make_aware(new_range_end, current_tz)

    response_data = {
        "range_start": new_range_start.strftime("%Y-%m-%d"),
        "range_end": new_range_end.strftime("%Y-%m-%d")
    }

    def large_stats_data(range_start, range_end):
        def get_objects(model_name, timestamp_field, date_field):
            """Perform timestamp based queries.

            This method receives a dynamic model name and performs a filter
            query. Later the results are grouped by day and prepared for the
            charting library.
            """
            args = '%s__range' % timestamp_field
            q = model_name.objects.filter(**{
                args: (range_start, range_end)
            })
            q = q.extra({timestamp_field: date_field})
            q = q.values(timestamp_field)
            q = q.annotate(created_count=Count('pk'))
            q = q.order_by(timestamp_field)

            data = []

            for obj in q:
                data.append([
                    time.mktime(time.strptime(
                        str(obj[timestamp_field]),
                        "%Y-%m-%d")) * 1000,
                    obj['created_count']
                ])

            return data

        comment_array = get_objects(Comment, "timestamp", "date(timestamp)")
        change_desc_array = get_objects(ChangeDescription, "timestamp",
                                        "date(timestamp)")
        review_array = get_objects(Review, "timestamp", "date(timestamp)")
        rr_array = get_objects(ReviewRequest, "time_added", "date(time_added)")

        return {
            'change_descriptions': change_desc_array,
            'comments': comment_array,
            'reviews': review_array,
            'review_requests': rr_array
        }

    stats_data = large_stats_data(new_range_start, new_range_end)

    return {
        "range": response_data,
        "activity_data": stats_data
    }


class ActivityGraphWidget(BaseAdminWidget):
    """Detailed database statistics graph widget.

    Shows the latest database activity for multiple models in the form of
    a graph that can be navigated by date.

    This widget shows a daily view of creation activity for a list of models.
    All displayed widget data is computed on demand, rather than up-front
    during creation of the widget.
    """

    widget_id = 'activity-graph-widget'
    name = _('Review Board Activity')
    js_model_class = 'RB.Admin.ServerActivityWidget'
    js_view_class = 'RB.Admin.ServerActivityWidgetView'
    css_classes = 'rb-c-admin-server-activity-widget'


def init_widgets():
    """Initialize the widgets subsystem.

    This will listen for events in order to manage the widget caches.
    """
    post_save.connect(_increment_sync_num, sender=Group)
    post_save.connect(_increment_sync_num, sender=Repository)
    post_delete.connect(_increment_sync_num, sender=Group)
    post_delete.connect(_increment_sync_num, sender=Repository)


#: The registry of available administration widgets.
admin_widgets_registry = AdminWidgetsRegistry()



"""A Review UI for text-based files."""

from __future__ import annotations

import logging
from typing import (Any, Iterator, List, Optional, Sequence, TYPE_CHECKING,
                    Union, cast)

from django.template.loader import render_to_string
from django.utils.encoding import force_bytes
from django.utils.safestring import mark_safe
from djblets.cache.backend import cache_memoize
from pygments import highlight
from pygments.lexers import (ClassNotFound, guess_lexer_for_filename,
                             TextLexer)

from reviewboard.attachments.mimetypes import TextMimetype
from reviewboard.attachments.models import FileAttachment
from reviewboard.diffviewer.chunk_generator import (NoWrapperHtmlFormatter,
                                                    RawDiffChunkGenerator)
from reviewboard.diffviewer.diffutils import get_chunks_in_range
from reviewboard.diffviewer.settings import DiffSettings
from reviewboard.reviews.models import FileAttachmentComment
from reviewboard.reviews.ui.base import (ReviewUI,
                                         SerializedComment,
                                         SerializedCommentBlocks)

if TYPE_CHECKING:
    from django.http import HttpRequest
    from pygments.lexers import Lexer
    from typelets.json import JSONDict


logger = logging.getLogger(__name__)


class SerializedTextComment(SerializedComment):
    """Serialized comment data for text review UIs.

    This must be kept in sync with the definitions in
    :file:`reviewboard/static/rb/js/reviews/models/commentData.ts`.

    Version Added:
        7.0
    """

    #: The starting line number of the comment.
    beginLineNum: int

    #: The ending line number of the comment.
    endLineNum: int

    #: The view mode of the document when the comment was made.
    #:
    #: This will be either "source" or "rendered".
    viewMode: str


class TextBasedReviewUI(ReviewUI[
    FileAttachment,
    FileAttachmentComment,
    SerializedTextComment
]):
    """A Review UI for text-based files.

    This renders the text file, applying syntax highlighting, and allows users
    to comment on one or more lines.
    """

    name = 'Text'
    object_key: str = 'text'
    supported_mimetypes = TextMimetype.supported_mimetypes
    template_name = 'reviews/ui/text.html'
    comment_thumbnail_template_name = 'reviews/ui/text_comment_thumbnail.html'
    can_render_text = False
    supports_diffing = True
    supports_file_attachments = True

    source_chunk_generator_cls = RawDiffChunkGenerator
    rendered_chunk_generator_cls = RawDiffChunkGenerator

    #: Extra classes to apply to the Review UI element.
    extra_css_classes: list[str] = []

    js_model_class: str = 'RB.TextBasedReviewable'
    js_view_class: str = 'RB.TextBasedReviewableView'

    def get_js_model_data(self) -> JSONDict:
        """Return data to pass to the JavaScript Model during instantiation.

        This data will be passed as attributes to the reviewable model
        when constructed.

        Returns:
            dict:
            The attributes to pass to the model.
        """
        data = super().get_js_model_data()
        data['hasRenderedView'] = self.can_render_text

        if self.can_render_text:
            data['viewMode'] = 'rendered'
        else:
            data['viewMode'] = 'source'

        return data

    def get_extra_context(
        self,
        request: HttpRequest,
    ) -> dict[str, Any]:
        """Return extra context to use when rendering the Review UI.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            dict:
            The context to provide to the template.
        """
        context = {}
        diff_type_mismatch = False

        if self.diff_against_obj:
            diff_against_review_ui = self.diff_against_obj.review_ui

            context.update({
                'diff_caption': self.diff_against_obj.caption,
                'diff_filename': self.diff_against_obj.filename,
                'diff_revision': self.diff_against_obj.attachment_revision,
            })

            if type(self) != type(diff_against_review_ui):
                diff_type_mismatch = True
            else:
                chunk_generator = self._get_source_diff_chunk_generator()
                context['source_chunks'] = chunk_generator.get_chunks()

                chunk_generator = self._get_rendered_diff_chunk_generator()
                context['rendered_chunks'] = chunk_generator.get_chunks()
        else:
            file_line_list = [
                mark_safe(line)
                for line in self.get_text_lines()
            ]

            rendered_line_list = [
                mark_safe(line)
                for line in self.get_rendered_lines()
            ]

            context.update({
                'text_lines': file_line_list,
                'rendered_lines': rendered_line_list,
            })

        if self.obj.attachment_history is not None:
            num_revisions = FileAttachment.objects.filter(
                attachment_history=self.obj.attachment_history).count()
        else:
            num_revisions = 1

        context.update({
            'filename': self.obj.filename,
            'revision': self.obj.attachment_revision,
            'is_diff': self.diff_against_obj is not None,
            'num_revisions': num_revisions,
            'diff_type_mismatch': diff_type_mismatch,
        })

        return context

    def get_text(self) -> str:
        """Return the file contents as a string.

        This will fetch the file and then cache it for future renders.

        Returns:
            str:
            The file contents as a text string.
        """
        return cast(
            str,
            cache_memoize('text-attachment-%d-string' % self.obj.pk,
                          self._get_text_uncached))

    def get_text_lines(self) -> Sequence[str]:
        """Return the file contents as syntax-highlighted lines.

        This will fetch the file, render it however appropriate for the review
        UI, and split it into reviewable lines. It will then cache it for
        future renders.
        """
        return cast(
            List[str],
            cache_memoize('text-attachment-%d-lines' % self.obj.pk,
                          lambda: list(self.generate_highlighted_text())))

    def get_rendered_lines(self) -> list[str]:
        """Return the file contents as a render, based on the raw text.

        If a subclass sets ``can_render_text = True`` and implements
        ``generate_render``, then this will render the contents in some
        specialized form, cache it as a list of lines, and return it.

        Returns:
            list of str:
            The rendered lines.
        """
        if self.can_render_text:
            return cast(
                List[str],
                cache_memoize('text-attachment-%d-rendered' % self.obj.pk,
                              lambda: list(self.generate_render())))
        else:
            return []

    def _get_text_uncached(self) -> str:
        """Return the text from the file.

        Returns:
            str:
            The text of the file.
        """
        self.obj.file.open()

        with self.obj.file as f:
            data = f.read()

        return data

    def generate_highlighted_text(self) -> list[str]:
        """Generate syntax-highlighted text for the file.

        This will render the text file to HTML, applying any syntax
        highlighting that's appropriate. The contents will be split into
        reviewable lines and will be cached for future renders.

        Returns:
            list of str:
            The syntax-highlighted text, split into lines.
        """
        assert self.obj.filename is not None

        data = self.get_text()

        lexer = self.get_source_lexer(self.obj.filename, data)
        lines = highlight(data, lexer, NoWrapperHtmlFormatter()).splitlines()

        return [
            '<pre>%s</pre>' % line
            for line in lines
        ]

    def get_source_lexer(
        self,
        filename: str,
        data: str,
    ) -> Lexer:
        """Return the lexer that should be used for the text.

        By default, this will attempt to guess the lexer based on the
        filename, falling back to a plain-text lexer.

        Subclasses can override this to choose a more specific lexer.

        Args:
            filename (str):
                The filename of the file.

            data (str):
                The contents of the file.

        Returns:
            pygments.lexers.Lexer:
            The lexer to use for highlighting the source.
        """
        try:
            return guess_lexer_for_filename(filename, data)
        except ClassNotFound:
            return TextLexer()

    def generate_render(self) -> Iterator[str]:
        """Generate a render of the text.

        By default, this won't do anything. Subclasses should override it
        to turn the raw text into some form of rendered content. For
        example, rendering Markdown.

        Yields:
            str:
            The rendered lines of content.
        """
        raise NotImplementedError

    def serialize_comments(
        self,
        comments: Sequence[FileAttachmentComment],
    ) -> SerializedCommentBlocks[SerializedTextComment]:
        """Serialize the comments for the file attachment.

        Args:
            comments (list of
                      reviewboard.reviews.models.FileAttachmentComment):
                The list of objects to serialize. This will be the result of
                :py:meth:`get_comments`.

        Returns:
            SerializedCommentBlocks:
            The serialized comments.
        """
        result: SerializedCommentBlocks[SerializedTextComment] = {}

        for comment in self.flat_serialized_comments(comments):
            try:
                key = f'{comment["beginLineNum"]}-{comment["endLineNum"]}'
            except KeyError:
                # It's possible this comment was made before the review UI
                # was provided, meaning it has no data. If this is the case,
                # ignore this particular comment, since it doesn't have a
                # line region.
                continue

            result.setdefault(key, []).append(comment)

        return result

    def get_comment_thumbnail(
        self,
        comment: FileAttachmentComment,
    ) -> Optional[str]:
        """Generate and return a thumbnail representing this comment.

        This will find the appropriate lines the comment applies to and
        return it as HTML suited for rendering in reviews.

        Args:
            comment (reviewboard.reviews.models.FileAttachmentComment):
                The comment to render the thumbnail for.

        Returns:
            str:
            The rendered comment thumbnail.
        """
        try:
            begin_line_num = int(comment.extra_data['beginLineNum'])
            end_line_num = int(comment.extra_data['endLineNum'])
            view_mode = comment.extra_data['viewMode']
        except (KeyError, ValueError):
            # This may be a comment from before we had review UIs. Or,
            # corrupted data. Either way, don't display anything.
            return None

        cache_key = 'text-review-ui-comment-thumbnail-%s-%s' % (
            self.obj.pk, comment.pk),

        return cast(
            Optional[str],
            cache_memoize(
                cache_key,
                lambda: self.render_comment_thumbnail(
                    comment, begin_line_num, end_line_num, view_mode)))

    def render_comment_thumbnail(
        self,
        comment: FileAttachmentComment,
        begin_line_num: int,
        end_line_num: int,
        view_mode: str,
    ) -> str:
        """Render the content of a comment thumbnail.

        This will, by default, call render() and then pull out the lines
        that were commented on.

        Subclasses can override to do more specialized thumbnail rendering.

        Args:
            comment (reviewboard.reviews.models.FileAttachmentComment):
                The comment to render the thumbnail for.

            begin_line_num (int):
                The starting line number for the comment.

            end_line_num (int):
                The end line number for the comment.

            view_mode (str):
                The view mode. One of ``source`` or ``rendered``.

        Returns:
            str:
            The rendered comment thumbnail.
        """
        if view_mode not in ('source', 'rendered'):
            logger.warning('Unexpected view mode "%s" when rendering '
                           'comment thumbnail.',
                           view_mode)
            return ''

        context = {
            'is_diff': self.diff_against_obj is not None,
            'review_ui': self,
            'revision': self.obj.attachment_revision,
        }

        if self.diff_against_obj:
            assert isinstance(self.diff_against_obj, FileAttachment)

            if view_mode == 'source':
                chunk_generator = self._get_source_diff_chunk_generator()
            elif view_mode == 'rendered':
                chunk_generator = self._get_rendered_diff_chunk_generator()

            chunks = get_chunks_in_range(chunk_generator.get_chunks(),
                                         begin_line_num,
                                         end_line_num - begin_line_num + 1)

            context.update({
                'chunks': chunks,
                'diff_revision': self.diff_against_obj.attachment_revision,
            })
        else:
            try:
                if view_mode == 'source':
                    lines = self.get_text_lines()
                elif view_mode == 'rendered':
                    lines = self.get_rendered_lines()
            except Exception as e:
                logger.error('Unable to generate text attachment comment '
                             'thumbnail for comment %s: %s',
                             comment, e)
                return ''

            # Grab only the lines we care about.
            #
            # The line numbers are stored 1-indexed, so normalize to 0.
            lines = lines[begin_line_num - 1:end_line_num]

            context['lines'] = [
                {
                    'line_num': begin_line_num + i,
                    'text': mark_safe(line),
                }
                for i, line in enumerate(lines)
            ]

        return render_to_string(
            template_name=self.comment_thumbnail_template_name,
            context=context)

    def get_comment_link_url(
        self,
        comment: FileAttachmentComment,
    ) -> str:
        """Return the URL to the file and line commented on.

        This will link to the correct file, view mode, and line for the
        given comment.

        Args:
            comment (reviewboard.reviews.models.FileAttachmentComment):
                The comment to link to.

        Returns:
            str:
            The URL to link the comment to.
        """
        base_url = super().get_comment_link_url(comment)

        try:
            begin_line_num = int(comment.extra_data['beginLineNum'])
            view_mode = comment.extra_data['viewMode']
        except (KeyError, ValueError):
            # This may be a comment from before we had review UIs. Or,
            # corrupted data. Either way, just return the default.
            return base_url

        return '%s#%s/line%s' % (base_url, view_mode, begin_line_num)

    def _get_diff_chunk_generator(
        self,
        chunk_generator_cls: type[RawDiffChunkGenerator],
        orig: Union[bytes, list[bytes]],
        modified: Union[bytes, list[bytes]],
    ) -> RawDiffChunkGenerator:
        """Return a chunk generator showing a diff for the text.

        The chunk generator will diff the text of this attachment against
        the text of the attachment being diffed against.

        This is used both for displaying the file attachment and
        rendering the thumbnail.

        Args:
            chunk_generator_cls (type):
                The chunk generator to instantiate. This should be a subclass
                of :py:class:`~reviewboard.diffviewer.chunk_generator
                .RawDiffChunkGenerator`.

            orig (bytes or list of bytes):
                The original file content to diff against.

            modified (bytes or list of bytes):
                The new file content.

        Returns:
            reviewboard.diffviewer.chunk_generator.RawDiffChunkGenerator:
            The chunk generator used to diff source or rendered text.
        """
        assert self.diff_against_obj is not None

        return chunk_generator_cls(
            old=orig,
            new=modified,
            orig_filename=self.obj.filename,
            modified_filename=self.diff_against_obj.filename,
            diff_settings=DiffSettings.create(request=self.request))

    def _get_source_diff_chunk_generator(self) -> RawDiffChunkGenerator:
        """Return a chunk generator for diffing source text.

        Returns:
            reviewboard.diffviewer.chunk_generator.RawDiffChunkGenerator:
            The chunk generator used to diff source text.
        """
        assert self.diff_against_obj is not None
        assert isinstance(self.diff_against_obj.review_ui, TextBasedReviewUI)

        return self._get_diff_chunk_generator(
            self.source_chunk_generator_cls,
            force_bytes(self.diff_against_obj.review_ui.get_text()),
            force_bytes(self.get_text()))

    def _get_rendered_diff_chunk_generator(self) -> RawDiffChunkGenerator:
        """Return a chunk generator for diffing rendered text.

        Returns:
            reviewboard.diffviewer.chunk_generator.RawDiffChunkGenerator:
            The chunk generator used to diff rendered text.
        """
        assert self.diff_against_obj is not None

        diff_against_review_ui = self.diff_against_obj.review_ui
        assert isinstance(diff_against_review_ui, TextBasedReviewUI)

        return self._get_diff_chunk_generator(
            self.rendered_chunk_generator_cls,
            [
                force_bytes(line)
                for line in diff_against_review_ui.get_rendered_lines()
            ],
            [
                force_bytes(line)
                for line in self.get_rendered_lines()
            ]
        )



"""Unit tests for reviewboard.admin.validation."""

from django.forms import ValidationError

from reviewboard.admin.validation import validate_bug_tracker
from reviewboard.testing.testcase import TestCase


class ValidationTests(TestCase):
    """Unit tests for reviewboard.admin.validation."""

    def test_validate_bug_tracker(self):
        """Testing validate_bug_tracker"""
        # Invalid - invalid format specification types
        with self.assertRaises(ValidationError):
            validate_bug_tracker('%20')

        with self.assertRaises(ValidationError):
            validate_bug_tracker('%d')

        # Invalid - too many format specification types
        with self.assertRaises(ValidationError):
            validate_bug_tracker('%s %s')

        # Invalid - no format specification types
        with self.assertRaises(ValidationError):
            validate_bug_tracker('www.a.com')

        # Valid - Escaped %'s, with a valid format specification type
        try:
            validate_bug_tracker('%%20%s')
        except ValidationError:
            self.fail('validate_bug_tracker() raised a ValidationError when '
                      'no error was expected.')



"""Remove index from FileAttachment.repo_path.

Version Added:
    2.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('FileAttachment', 'repo_path', initial=None, db_index=False)
]



"""Allow Repository.extra_data to be NULL.

Version Added:
    1.6.7
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Repository', 'extra_data', initial=None, null=True),
]



"""Unit tests for reviewboard.actions.base.BaseGroupAction.

Version Added:
    8.0
"""

from __future__ import annotations

from django.template import Context
from django.utils.safestring import SafeString

from reviewboard.actions.renderers import DefaultActionGroupRenderer
from reviewboard.actions.tests.base import (
    TestActionsRegistry,
    TestGroupAction,
    TestGroupItemAction1,
    TestGroupItemAction2,
    TestGroupItemAction3,
)
from reviewboard.testing import TestCase


class BaseGroupActionTests(TestCase):
    """Unit tests for BaseGroupAction.

    Version Added:
        8.0
    """

    ######################
    # Instance variables #
    ######################

    #: A BaseGroupAction instance for testing.
    group_action: TestGroupAction

    #: Item 1's action, for testing.
    item1_action: TestGroupItemAction1

    #: Item 2's action, for testing.
    item2_action: TestGroupItemAction2

    #: Item 3's action, for testing.
    item3_action: TestGroupItemAction3

    #: An empty actions registry for testing.
    registry: TestActionsRegistry

    def setUp(self) -> None:
        """Set up state for a test."""
        super().setUp()

        group_action = TestGroupAction()
        item1_action = TestGroupItemAction1()
        item2_action = TestGroupItemAction2()
        item3_action = TestGroupItemAction3()

        registry = TestActionsRegistry()
        registry.register(group_action)
        registry.register(item1_action)
        registry.register(item2_action)
        registry.register(item3_action)

        self.registry = registry
        self.group_action = group_action
        self.item1_action = item1_action
        self.item2_action = item2_action
        self.item3_action = item3_action

    def tearDown(self) -> None:
        """Tear down state for a test."""

        del self.group_action
        del self.item1_action
        del self.item2_action
        del self.item3_action
        del self.registry

        super().tearDown()

    def test_get_extra_context(self) -> None:
        """Testing BaseGroupAction.get_extra_context"""
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertEqual(
            self.group_action.get_extra_context(request=request,
                                                context=context),
            {
                'id': 'group-action',
                'label': 'Test Group',
                'url': '#',
                'verbose_label': None,
                'visible': True,
            })

    def test_js_model_data(self) -> None:
        """Testing BaseGroupAction.js_model_data"""
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        self.assertEqual(
            self.group_action.get_js_model_data(context=context),
            {
                'id': 'group-action',
                'children': {
                    'review-request': [
                        'group-item-2-action',
                        'group-item-1-action',
                        'group-item-3-action',
                    ],
                },
                'label': 'Test Group',
                'url': '#',
                'visible': True,
            })

    def test_render_with_default_renderer(self) -> None:
        """Testing BaseGroupAction.render with default renderer"""
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        action = self.group_action

        html = action.render(
            request=request,
            context=context,
            fallback_renderer=DefaultActionGroupRenderer,
            placement=action.get_placement('review-request'),
        )

        self.assertIsInstance(html, SafeString)
        self.assertHTMLEqual(
            html,
            """
            <li class="rb-c-actions__action"
                role="group">
             <a id="action-review-request-group-item-1-action"
                role="button"
                href="#">
              Group Item 1
             </a>
             <a id="action-review-request-group-item-2-action"
                role="button"
                href="#">
              Group Item 2
             </a>
             <a id="action-review-request-group-item-3-action"
                role="button"
                href="#">
              Group Item 3
             </a>
            </li>
            """)

    def test_render_js_with_default_renderer(self) -> None:
        """Testing BaseGroupAction.render_js with default renderer"""
        request = self.create_http_request()
        context = Context({
            'request': request,
        })

        action = self.group_action

        js = action.render_js(
            request=request,
            context=context,
            fallback_renderer=DefaultActionGroupRenderer,
            placement=action.get_placement('review-request'),
        )

        self.assertIsInstance(js, SafeString)
        self.assertHTMLEqual(
            js,
            """
            page.addActionView(new RB.Actions.ActionView({
                "attachmentPointID": "review-request",
                el: $('#action-review-request-group-action'),
                model: page.getAction("group-action"),
            }));
            """)



"""Unit tests for reviewboard.reviews.ui.text.MarkdownReviewUI."""

from django.test.client import RequestFactory

from reviewboard.reviews.ui.markdownui import MarkdownReviewUI
from reviewboard.testing import TestCase


class MarkdownReviewUITests(TestCase):
    """Unit tests for reviewboard.reviews.ui.text.MarkdownReviewUI."""

    fixtures = ['test_users']

    def setUp(self):
        super(MarkdownReviewUITests, self).setUp()

        self.review_request = self.create_review_request()
        self.attachment_history = self.create_file_attachment_history(
            self.review_request)
        self.attachment = self.create_file_attachment(
            self.review_request,
            attachment_history=self.attachment_history,
            attachment_revision=0,
            caption='Revision 1 caption',
            orig_filename='revision1.md',
            mimetype='text/x-markdown',
            file_content=b'This is revision *1*.')
        self.review = self.create_review(self.review_request)

    def test_get_extra_context(self):
        """Testing MarkdownReviewUI.get_extra_context"""
        new_attachment = self.create_file_attachment(
            self.review_request,
            attachment_history=self.attachment_history,
            attachment_revision=1,
            caption='Revision 2 caption',
            orig_filename='revision2.md',
            mimetype='text/x-markdown',
            file_content=b'And **this** is revision 2.')

        review_ui = MarkdownReviewUI(review_request=self.review_request,
                                     obj=new_attachment)

        request = RequestFactory().get('/')
        extra_context = review_ui.get_extra_context(request)

        self.assertFalse(extra_context['is_diff'])
        self.assertFalse(extra_context['diff_type_mismatch'])
        self.assertEqual(extra_context['filename'], 'revision2.md')
        self.assertEqual(extra_context['revision'], 1)
        self.assertEqual(extra_context['num_revisions'], 2)
        self.assertNotIn('diff_caption', extra_context)
        self.assertNotIn('diff_filename', extra_context)
        self.assertNotIn('diff_revision', extra_context)
        self.assertNotIn('source_chunks', extra_context)
        self.assertNotIn('rendered_chunks', extra_context)
        self.assertEqual(extra_context['text_lines'],
                         ['<pre>And **this** is revision 2.</pre>'])
        self.assertEqual(extra_context['rendered_lines'],
                         ['<p>And <strong>this</strong> is revision 2.</p>'])

    def test_get_extra_context_with_diff(self):
        """Testing MarkdownReviewUI.get_extra_context with diff_against_obj"""
        new_attachment = self.create_file_attachment(
            self.review_request,
            attachment_history=self.attachment_history,
            attachment_revision=1,
            caption='Revision 2 caption',
            orig_filename='revision2.md',
            mimetype='text/x-markdown',
            file_content=b'And **this** is revision 2.')

        review_ui = MarkdownReviewUI(review_request=self.review_request,
                                     obj=new_attachment)
        review_ui.set_diff_against(self.attachment)

        request = RequestFactory().get('/')
        extra_context = review_ui.get_extra_context(request)

        self.assertTrue(extra_context['is_diff'])
        self.assertFalse(extra_context['diff_type_mismatch'])
        self.assertEqual(extra_context['filename'], 'revision2.md')
        self.assertEqual(extra_context['revision'], 1)
        self.assertEqual(extra_context['num_revisions'], 2)
        self.assertEqual(extra_context['diff_caption'], 'Revision 1 caption')
        self.assertEqual(extra_context['diff_filename'], 'revision1.md')
        self.assertEqual(extra_context['diff_revision'], 0)
        self.assertEqual(
            list(extra_context['source_chunks']),
            [
                {
                    'change': 'replace',
                    'collapsable': False,
                    'index': 0,
                    'lines': [
                        [
                            1,
                            1,
                            'This is revision <span class="ge">*1*</span>.',
                            [(0, 1), (4, 4), (17, 20)],
                            1,
                            'And <span class="gs">**this**</span> is '
                            'revision 2.',
                            [(0, 7), (10, 12), (25, 26)],
                            False,
                        ],
                    ],
                    'meta': {
                        'left_headers': [],
                        'right_headers': [],
                        'whitespace_chunk': False,
                        'whitespace_lines': [],
                    },
                    'numlines': 1,
                },
            ])
        self.assertEqual(
            list(extra_context['rendered_chunks']),
            [
                {
                    'change': 'replace',
                    'collapsable': False,
                    'index': 0,
                    'lines': [
                        [
                            1,
                            1,
                            '<p>This is revision <em>1</em>.</p>',
                            [(0, 1), (17, 18)],
                            1,
                            '<p>And <strong>this</strong> is revision 2.</p>',
                            [(0, 5), (21, 22)],
                            False,
                        ],
                    ],
                    'meta': {
                        'left_headers': [],
                        'right_headers': [],
                        'whitespace_chunk': False,
                        'whitespace_lines': [],
                    },
                    'numlines': 1,
                },
            ])
        self.assertNotIn('text_lines', extra_context)
        self.assertNotIn('rendered_lines', extra_context)



"""Unit tests for reviewboard.reviews.manager.StatusUpdateManager.

Version Added:
    5.0.3
"""

from __future__ import annotations

from reviewboard.integrations.base import Integration, get_integration_manager
from reviewboard.integrations.models import IntegrationConfig
from reviewboard.reviews.models import StatusUpdate
from reviewboard.testing import TestCase


class MyIntegration(Integration):
    integration_id = 'my-integration'
    name = 'My Integration'


class StatusUpdateManagerTests(TestCase):
    """Unit tests for reviewboard.reviews.manager.StatusUpdateManager.

    Version Added:
        5.0.3
    """

    fixtures = ['test_users']

    def test_create_for_integration(self) -> None:
        """Testing StatusUpdateManager.create_for_integration"""
        self._run_create_for_integration_test()

    def test_create_for_integration_with_description(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with description=
        """
        self._run_create_for_integration_test(
            description='my custom description...',
            expected_description='my custom description...')

    def test_create_for_integration_with_state(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with state=
        """
        self._run_create_for_integration_test(
            state=StatusUpdate.DONE_FAILURE,
            expected_state=StatusUpdate.DONE_FAILURE)

    def test_create_for_integration_with_service_id(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with service_id=
        """
        self._run_create_for_integration_test(
            service_id='xxx-my-service',
            expected_service_id='xxx-my-service')

    def test_create_for_integration_with_run_manually(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with
        IntegrationConfig.run_manually
        """
        self._run_create_for_integration_test(
            config_run_manually=True,
            expected_state=StatusUpdate.NOT_YET_RUN,
            expected_description='waiting to run.')

    def test_create_for_integration_with_run_manually_and_description(
        self,
    ) -> None:
        """Testing StatusUpdateManager.create_for_integration with
        IntegrationConfig.run_manually and description=
        """
        self._run_create_for_integration_test(
            config_run_manually=True,
            description='my custom description...',
            expected_state=StatusUpdate.NOT_YET_RUN,
            expected_description='my custom description...')

    def test_create_for_integration_with_run_manually_and_state(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with
        IntegrationConfig.run_manually and state=
        """
        self._run_create_for_integration_test(
            config_run_manually=True,
            state=StatusUpdate.DONE_FAILURE,
            expected_state=StatusUpdate.DONE_FAILURE,
            expected_description='waiting to run.')

    def test_create_for_integration_with_can_retry(self) -> None:
        """Testing StatusUpdateManager.create_for_integration with
        can_retry=True
        """
        self._run_create_for_integration_test(can_retry=True)

    def _run_create_for_integration_test(
        self,
        *,
        config_run_manually: bool = False,
        expected_state: str = StatusUpdate.PENDING,
        expected_service_id: str = 'my-integration',
        expected_summary: str = 'My Integration',
        expected_description: str = 'starting...',
        **kwargs,
    ) -> None:
        """Run a test for create_for_integration.

        Args:
            config_run_manually (bool, optional):
                Whether to set ``run_manually=True`` in the configuration.

            expected_state (str, optional):
                The expected Status Update state.

            expected_service_id (str, optional):
                The expected Status Update service ID.

            expected_summary (str, optional):
                The expected Status Update summary.

            expected_description (str, optional):
                The expected Status Update description.

            **kwargs (dict, optional):
                Keyword arguments to pass to
                :py:meth:`StatusUpdate.create_for_integration()
                <reviewboard.reviews.managers.StatusUpdate.
                create_for_integration>`.

        Raises:
            AssertionError:
                An expectation failed in the test.
        """
        config = IntegrationConfig.objects.create(
            integration_id=MyIntegration.integration_id,
            name='my-config')

        if config_run_manually:
            config.set('run_manually', True)

        bot_user = self.create_user(username='bot')
        review_request = self.create_review_request()

        integration_mgr = get_integration_manager()
        integration_mgr.register_integration_class(MyIntegration)

        try:
            integration = integration_mgr.get_integration(
                MyIntegration.integration_id)

            status_update = StatusUpdate.objects.create_for_integration(
                integration,
                config=config,
                user=bot_user,
                review_request=review_request,
                **kwargs)
        finally:
            integration_mgr.unregister_integration_class(MyIntegration)

        self.assertEqual(status_update.state, expected_state)
        self.assertEqual(status_update.service_id, expected_service_id)
        self.assertEqual(status_update.summary, expected_summary)
        self.assertEqual(status_update.description, expected_description)

        self.assertEqual(status_update.integration_config, config)
        self.assertEqual(status_update.user, bot_user)
        self.assertEqual(status_update.review_request, review_request)
        self.assertEqual(status_update.extra_data, {
            '__integration_config_id': config.pk,
            'can_retry': kwargs.get('can_retry', False),
        })



"""Unit tests for reviewboard.admin.form_widgets.RelatedUserWidget."""

from django import forms
from django.contrib.auth.models import User

from reviewboard.admin.form_widgets import RelatedUserWidget
from reviewboard.testing.testcase import TestCase


class TestForm(forms.Form):
    """A Test Form with a field that contains a RelatedUserWidget."""
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        label=('Default users'),
        required=False,
        widget=RelatedUserWidget())


class LocalSiteTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedUserWidget.

    The RelatedUserWidget is defined to have a local_site_name.
    """
    my_multiselect_field = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        label=('Default users'),
        required=False,
        widget=RelatedUserWidget(local_site_name='supertest'))


class SingleValueTestForm(forms.Form):
    """A Test Form with a field that contains a RelatedUserWidget.

    The RelatedUserWidget is defined as setting multivalued to False.
    """
    my_select_field = forms.ModelMultipleChoiceField(
        queryset=User.objects.filter(is_active=True),
        label=('Default users'),
        required=False,
        widget=RelatedUserWidget(multivalued=False))


class RelatedUserWidgetTests(TestCase):
    """Unit tests for RelatedUserWidget."""

    fixtures = ['test_users']

    def test_render_empty(self):
        """Testing RelatedUserWidget.render with no initial data"""
        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default users',
            [],
            {'id': 'default-users'})

        self.assertHTMLEqual(
            """<input id="default-users" name="Default users" type="hidden" />

            <script>
            $(function() {
                var view = new RB.RelatedUserSelectorView({
                    $input: $('#default\\u002Dusers'),
                    initialOptions: [],

                    useAvatars: true,
                    multivalued: true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_data(self):
        """Testing RelatedUserWidget.render with initial data"""
        my_form = TestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default users',
            [1, 2, 3],
            {'id': 'default-users'})

        self.assertHTMLEqual(
            """<input id="default-users" name="Default users"
            type="hidden" value="1,2,3" />

            <script>
            $(function() {
                var view = new RB.RelatedUserSelectorView({
                    $input: $('#default\\u002Dusers'),
                    initialOptions: [{"avatarHTML": "\\u003Cimg
                    src=\\"https://secure.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=20\\u0026d=mm\\"
                    alt=\\"admin\\" width=\\"20\\" height=\\"20\\"
                    srcset=\\"https://secure.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=20\\u0026d=mm 1x,
                    https://secure.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=40\\u0026d=mm 2x,
                    https://secure.gravatar.com/avatar/e64c7d89f26bd1972efa854d13d7dd61?s=60\\u0026d=mm 3x\\"
                    class=\\"avatar djblets-o-avatar\\"\\u003E\\n",
                    "fullname": "Admin User",
                    "id": 1,
                    "username": "admin"},
                    {"avatarHTML": "\\u003Cimg
                    src=\\"https://secure.gravatar.com/avatar/b0f1ae4342591db2695fb11313114b3e?s=20\\u0026d=mm\\"
                    alt=\\"doc\\" width=\\"20\\" height=\\"20\\"
                    srcset=\\"https://secure.gravatar.com/avatar/b0f1ae4342591db2695fb11313114b3e?s=20\\u0026d=mm 1x,
                    https://secure.gravatar.com/avatar/b0f1ae4342591db2695fb11313114b3e?s=40\\u0026d=mm 2x,
                    https://secure.gravatar.com/avatar/b0f1ae4342591db2695fb11313114b3e?s=60\\u0026d=mm 3x\\"
                    class=\\"avatar djblets-o-avatar\\"\\u003E\\n",
                    "fullname": "Doc Dwarf",
                    "id": 2,
                    "username": "doc"},
                    {"avatarHTML": "\\u003Cimg
                    src=\\"https://secure.gravatar.com/avatar/1a0098e6600792ea4f714aa205bf3f2b?s=20\\u0026d=mm\\"
                    alt=\\"dopey\\" width=\\"20\\" height=\\"20\\"
                    srcset=\\"https://secure.gravatar.com/avatar/1a0098e6600792ea4f714aa205bf3f2b?s=20\\u0026d=mm 1x,
                    https://secure.gravatar.com/avatar/1a0098e6600792ea4f714aa205bf3f2b?s=40\\u0026d=mm 2x,
                    https://secure.gravatar.com/avatar/1a0098e6600792ea4f714aa205bf3f2b?s=60\\u0026d=mm 3x\\"
                    class=\\"avatar djblets-o-avatar\\"\\u003E\\n",
                    "fullname": "Dopey Dwarf",
                    "id": 3,
                    "username": "dopey"}],

                    useAvatars: true,
                    multivalued: true
                }).render();
            });
            </script>""",
            html)

    def test_render_with_local_site(self):
        """Testing RelatedUserWidget.render with a local site defined"""
        my_form = LocalSiteTestForm()
        html = my_form.fields['my_multiselect_field'].widget.render(
            'Default users',
            [],
            {'id': 'default-users'})

        self.assertIn("localSitePrefix: 's/supertest/',", html)

    def test_value_from_datadict(self):
        """Testing RelatedUserWidget.value_from_datadict"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'people': ['1', '2']},
                {},
                'people'))

        self.assertEqual(value, ['1', '2'])

    def test_value_from_datadict_single_value(self):
        """Testing RelatedUserWidget.value_from_datadict with a single value"""
        my_form = SingleValueTestForm()
        value = (
            my_form.fields['my_select_field']
            .widget
            .value_from_datadict(
                {'people': ['1']},
                {},
                'people'))

        self.assertEqual(value, ['1'])

    def test_value_from_datadict_with_no_data(self):
        """Testing RelatedUserWidget.value_from_datadict with no data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {'people': []},
                {},
                'people'))
        self.assertEqual(value, [])

    def test_value_from_datadict_with_missing_data(self):
        """Testing RelatedUserWidget.value_from_datadict with missing data"""
        my_form = TestForm()
        value = (
            my_form.fields['my_multiselect_field']
            .widget
            .value_from_datadict(
                {},
                {},
                'people'))

        self.assertIsNone(value)



"""Support for Splat as a bug tracker."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from django import forms
from django.utils.translation import gettext_lazy as _

from reviewboard.hostingsvcs.base.bug_tracker import BaseBugTracker
from reviewboard.hostingsvcs.base.forms import BaseHostingServiceRepositoryForm
from reviewboard.hostingsvcs.base.hosting_service import BaseHostingService

if TYPE_CHECKING:
    from reviewboard.hostingsvcs.base.bug_tracker import BugInfo
    from reviewboard.scmtools.models import Repository


logger = logging.getLogger(__name__)


class SplatForm(BaseHostingServiceRepositoryForm):
    """The Splat bug tracker configuration form."""

    splat_org_name = forms.SlugField(
        label=_('Splat Organization Name'),
        max_length=64,
        required=True,
        widget=forms.TextInput(attrs={'size': 60}))


class Splat(BaseHostingService, BaseBugTracker):
    """The Splat bug tracker.

    Splat is a SaaS bugtracker hosted at https://hellosplat.com. It is owned
    and run by Beanbag, Inc. and is used as the official bug tracker for
    Review Board.
    """

    name = 'Splat'
    hosting_service_id = 'splat'
    form = SplatForm
    bug_tracker_field = \
        'https://hellosplat.com/s/%(splat_org_name)s/tickets/%%s/'
    supports_bug_trackers = True

    def get_bug_info_uncached(
        self,
        repository: Repository,
        bug_id: str,
    ) -> BugInfo:
        """Return the information for the specified bug.

        Args:
            repository (reviewboard.scmtools.models.Repository):
                The repository object.

            bug_id (str):
                The ID of the bug to fetch.

        Returns:
            reviewboard.hostingsvcs.base.bug_tracker.BugInfo:
            Information about the bug.
        """
        result: BugInfo = {
            'summary': '',
            'description': '',
            'description_text_format': '',
            'status': '',
        }

        org_name = repository.extra_data['bug_tracker-splat_org_name']
        url = (
            f'https://hellosplat.com/api/orgs/{org_name}/tickets/{bug_id}/'
            f'?only-fields=status,summary,text,text_format'
        )

        try:
            rsp = self.client.json_get(url)[0]
            ticket = rsp['ticket']
        except Exception as e:
            logger.warning('Unable to fetch Splat data from %s: %s',
                           url, e, exc_info=True)
        else:
            text = ticket['text']

            # Normalize the text format. For Splat, this is going to look
            # a bit silly, but we don't want to make the assumption in code
            # that we can just pass through the text format here.
            text_format = {
                'plain': 'plain',
                'markdown': 'markdown',
                'html': 'html',
            }.get(ticket['text_format'], 'plain')

            result = {
                'description': text,
                'description_text_format': text_format,
                'status': ticket['status'],
                'summary': ticket['summary'],
            }

        return result



"""Limit Repository.path, mirror_path, and raw_file_url to 255 chars.

The old max_lengths were 256 characters.

Version Added:
    1.5
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('Repository', 'path', initial=None, max_length=255),
    ChangeField('Repository', 'mirror_path', initial=None, max_length=255),
    ChangeField('Repository', 'raw_file_url', initial=None, max_length=255)
]



"""Unit tests for SAML views."""

from __future__ import annotations

from typing import ClassVar, TYPE_CHECKING
from xml.etree import ElementTree

import kgb
from django.contrib.auth.models import User
from django.core.cache import cache
from django.urls import reverse
from djblets.cache.backend import make_cache_key

from reviewboard.accounts.sso.backends.saml.views import (SAMLACSView,
                                                          SAMLLinkUserView,
                                                          SAMLSLSView)
from reviewboard.testing import TestCase

if TYPE_CHECKING:
    from typelets.json import JSONDict


VALID_CERT = """-----BEGIN CERTIFICATE-----
MIICZjCCAc+gAwIBAgIBADANBgkqhkiG9w0BAQ0FADBQMQswCQYDVQQGEwJ1czEL
MAkGA1UECAwCQ0ExFjAUBgNVBAoMDUJlYW5iYWcsIEluYy4xHDAaBgNVBAMME2h0
dHBzOi8vZXhhbXBsZS5jb20wHhcNMjIwNTA2MTU0NjI1WhcNMjMwNTA2MTU0NjI1
WjBQMQswCQYDVQQGEwJ1czELMAkGA1UECAwCQ0ExFjAUBgNVBAoMDUJlYW5iYWcs
IEluYy4xHDAaBgNVBAMME2h0dHBzOi8vZXhhbXBsZS5jb20wgZ8wDQYJKoZIhvcN
AQEBBQADgY0AMIGJAoGBANCsbj4mvUiQERBy80R7yqA6hU3FMM4siC2UcUS3ltFF
grkVOAPr+zUnrdadmAiTH35AB94oMzf0Qh8OJCr7wG5JQm686TRkVm2xUxhJUcoq
7LjBTKeEXBcrEzdNlagFXxHUSz5bPSdwDt/zbOfe+9RZKeb4FggFCEYw/mi69+Dx
AgMBAAGjUDBOMB0GA1UdDgQWBBS4cP9Y+IM7ZHZChUDdx68QExTZUDAfBgNVHSME
GDAWgBS4cP9Y+IM7ZHZChUDdx68QExTZUDAMBgNVHRMEBTADAQH/MA0GCSqGSIb3
DQEBDQUAA4GBALht5/NfJU+GxYfQKiGkZ4Ih/T/48rzXAT7/7f61s7w72UR2S5e2
WsR7/JPkZ5+u5mCgmABjNcd9NzaBM2RfSrrurwbjXMQ8nb/+REvhXXJ4STsS48y5
bef2JtIf7mGDw8/KsUrAA2jEIpCedToGyQxyE6GdN5b69ITWvyAemnIM
-----END CERTIFICATE-----"""


class SAMLMetadataViewTests(TestCase):
    """Unit tests for SAMLMetadataView."""

    def test_metadata_view(self) -> None:
        """Testing SAMLMetadataView"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_verification_cert': VALID_CERT,
        }

        with self.siteconfig_settings(settings):
            url = reverse('sso:saml:metadata', kwargs={'backend_id': 'saml'})
            rsp = self.client.get(url)

            self.assertEqual(rsp.status_code, 200)

            root = ElementTree.fromstring(rsp.content)

            namespaces = {'md': 'urn:oasis:names:tc:SAML:2.0:metadata'}

            descriptor = root.find('md:SPSSODescriptor', namespaces)
            assert descriptor is not None

            sls = descriptor.find('md:SingleLogoutService', namespaces)
            assert sls is not None

            self.assertEqual(
                sls.get('Binding'),
                'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Redirect')
            self.assertEqual(
                sls.get('Location'),
                'http://example.com/account/sso/saml/sls/')

            acs = descriptor.find('md:AssertionConsumerService', namespaces)
            assert acs is not None

            self.assertEqual(
                acs.get('Binding'),
                'urn:oasis:names:tc:SAML:2.0:bindings:HTTP-POST')
            self.assertEqual(
                acs.get('Location'),
                'http://example.com/account/sso/saml/acs/')


class SAMLLinkUserViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for SAMLLinkUserView."""

    fixtures: ClassVar[list[str]] = ['test_users']

    def test_get_link_user_existing_account(self) -> None:
        """Testing SAMLLinkUserView form render with existing account"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc@example.com',
                },
            }
            session.save()

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.get(url)

            self.assertEqual(rsp.status_code, 200)

            context = rsp.context
            self.assertEqual(context['user'].username, 'doc')
            self.assertEqual(context['mode'],
                             SAMLLinkUserView.Mode.CONNECT_EXISTING_ACCOUNT)

    def test_get_link_user_existing_account_email_match(self) -> None:
        """Testing SAMLLinkUserView form render with existing account matching
        email address
        """
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc2',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc@example.com',
                },
            }
            session.save()

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.get(url)

            self.assertEqual(rsp.status_code, 200)

            context = rsp.context
            self.assertEqual(context['user'].username, 'doc')
            self.assertEqual(context['mode'],
                             SAMLLinkUserView.Mode.CONNECT_EXISTING_ACCOUNT)

    def test_get_link_user_existing_account_email_username_match(self) -> None:
        """Testing SAMLLinkUserView form render with existing account matching
        username from email address
        """
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc2',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc@example.org',
                },
            }
            session.save()

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.get(url)

            self.assertEqual(rsp.status_code, 200)

            context = rsp.context
            self.assertEqual(context['user'].username, 'doc')
            self.assertEqual(context['mode'],
                             SAMLLinkUserView.Mode.CONNECT_EXISTING_ACCOUNT)

    def test_get_link_user_no_match(self) -> None:
        """Testing SAMLLinkUserView form render with no match"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc2',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc2@example.org',
                },
            }
            session.save()

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.get(url)

            self.assertEqual(rsp.status_code, 200)

            context = rsp.context
            self.assertEqual(context['user'], None)
            self.assertEqual(context['mode'],
                             SAMLLinkUserView.Mode.PROVISION)

    def test_post_link_user_login(self) -> None:
        """Testing SAMLLinkUserView form POST with login"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc2',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc@example.com',
                },
            }
            session.save()

            user = User.objects.get(username='doc')
            self.assertFalse(user.linked_accounts.exists())

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.post(url, {
                'username': 'doc',
                'password': 'doc',
                'provision': '0',
            })

            self.assertEqual(rsp.status_code, 302)

            linked_accounts = list(user.linked_accounts.all())

            self.assertEqual(len(linked_accounts), 1)
            linked_account = linked_accounts[0]
            self.assertEqual(linked_account.service_id, 'sso:saml')
            self.assertEqual(linked_account.service_user_id, 'doc2')

    def test_post_link_user_provision(self) -> None:
        """Testing SAMLLinkUserView form POST with provision"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'sleepy',
                    'first_name': 'Sleepy',
                    'last_name': 'Dwarf',
                    'email': 'sleepy@example.com',
                },
            }
            session.save()

            self.assertFalse(User.objects.filter(username='sleepy').exists())

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.post(url, {
                'username': '',
                'password': '',
                'provision': '1',
            })

            self.assertEqual(rsp.status_code, 302)

            user = User.objects.get(username='sleepy')
            self.assertEqual(user.first_name, 'Sleepy')
            self.assertEqual(user.last_name, 'Dwarf')
            self.assertEqual(user.email, 'sleepy@example.com')

            linked_accounts = list(user.linked_accounts.all())

            self.assertEqual(len(linked_accounts), 1)
            linked_account = linked_accounts[0]
            self.assertEqual(linked_account.service_id, 'sso:saml')
            self.assertEqual(linked_account.service_user_id, 'sleepy')

    def test_post_link_user_provision_disabled(self) -> None:
        """Testing SAMLLinkUserView form POST with provision when provisioning
        is disabled
        """
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
            'saml_automatically_provision_users': False,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'sleepy',
                    'first_name': 'Sleepy',
                    'last_name': 'Dwarf',
                    'email': 'sleepy@example.com',
                },
            }
            session.save()

            self.assertFalse(User.objects.filter(username='sleepy').exists())

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.post(url, {
                'username': '',
                'password': '',
                'provision': '1',
            })

            self.assertEqual(rsp.status_code, 200)
            self.assertFalse(User.objects.filter(username='sleepy').exists())

    def test_post_link_user_inactive(self) -> None:
        """Testing SAMLLinkUserView form POST with an inactive user"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'doc2',
                    'first_name': 'Doc',
                    'last_name': 'Dwarf',
                    'email': 'doc@example.com',
                },
            }
            session.save()

            user = User.objects.get(username='doc')
            user.is_active = False
            user.save(update_fields=['is_active'])
            self.assertFalse(user.linked_accounts.exists())

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.post(url, {
                'username': 'doc',
                'password': 'doc',
                'provision': '0',
            })

            self.assertEqual(rsp.status_code, 200)

            form = rsp.context['form']
            self.assertFalse(form.is_valid())
            self.assertEqual(form.errors['__all__'],
                             ['This account is inactive.'])

            linked_accounts = list(user.linked_accounts.all())

            self.assertEqual(len(linked_accounts), 0)

    def test_post_link_user_invalid_mode(self) -> None:
        """Testing SAMLLinkUserView form POST with invalid mode"""
        settings: JSONDict = {
            'saml_enabled': True,
            'saml_require_login_to_link': True,
        }

        with self.siteconfig_settings(settings):
            session = self.client.session
            session['sso'] = {
                'user_data': {
                    'id': 'sleepy',
                    'first_name': 'Sleepy',
                    'last_name': 'Dwarf',
                    'email': 'sleepy@example.com',
                },
            }
            session.save()

            self.assertFalse(User.objects.filter(username='sleepy').exists())

            url = reverse('sso:saml:link-user', kwargs={'backend_id': 'saml'})
            rsp = self.client.post(url, {
                'username': '',
                'password': '',
                'provision': '1',
                'mode': 'invalid',
            })

            self.assertEqual(rsp.status_code, 302)

            user = User.objects.get(username='sleepy')
            self.assertEqual(user.first_name, 'Sleepy')
            self.assertEqual(user.last_name, 'Dwarf')
            self.assertEqual(user.email, 'sleepy@example.com')

            linked_accounts = list(user.linked_accounts.all())

            self.assertEqual(len(linked_accounts), 1)
            linked_account = linked_accounts[0]
            self.assertEqual(linked_account.service_id, 'sso:saml')
            self.assertEqual(linked_account.service_user_id, 'sleepy')


class SAMLACSViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for SAMLACSView."""

    fixtures: ClassVar[list[str]] = ['test_users']

    def setUp(self) -> None:
        """Set up the test case."""
        class FakeAuth:
            def process_response(*args, **kwargs):
                pass

            def get_last_message_id(self):
                return 'message-id'

            def get_last_assertion_id(self):
                return 'assertion-id'

            def get_last_error_reason(self):
                return None

            def get_nameid(self):
                return 'username'

            def get_attribute(self, attr):
                return ['a']

            def get_attributes(self):
                return {}

            def get_session_index(self):
                return 1

        fake_auth = FakeAuth()

        self.spy_on(SAMLACSView.get_saml_auth,
                    op=kgb.SpyOpReturn(fake_auth),
                    owner=SAMLACSView)

        super().setUp()

    def tearDown(self) -> None:
        """Tear down the test case."""
        cache.delete(make_cache_key('saml_replay_id_message-id'))

        super().tearDown()

    def test_post_assertion_login(self) -> None:
        """Testing SAMLACSView POST log in to existing linked account"""
        settings: JSONDict = {
            'saml_enabled': True,
        }

        user = User.objects.get(username='doc')
        user.linked_accounts.create(
            service_id='sso:saml',
            service_user_id='username')

        self.assertNotIn('_auth_user_id', self.client.session)

        with self.siteconfig_settings(settings):
            url = reverse('sso:saml:acs', kwargs={'backend_id': 'saml'})

            rsp = self.client.post(
                path=url,
                data={},
                HTTP_HOST='localhost')
            self.assertEqual(rsp.status_code, 302)

            # Make sure we've logged in as the expected user.
            self.assertEqual(int(self.client.session['_auth_user_id']),
                             user.pk)

    def test_post_assertion_login_with_inactive_user(self) -> None:
        """Testing SAMLACSView POST log in to existing linked account with an
        inactive user
        """
        settings: JSONDict = {
            'saml_enabled': True,
        }

        user = User.objects.get(username='doc')
        user.is_active = False
        user.save(update_fields=['is_active'])
        user.linked_accounts.create(
            service_id='sso:saml',
            service_user_id='username')

        self.assertNotIn('_auth_user_id', self.client.session)

        with self.siteconfig_settings(settings):
            url = reverse('sso:saml:acs', kwargs={'backend_id': 'saml'})

            rsp = self.client.post(
                path=url,
                data={},
                HTTP_HOST='localhost')
            self.assertEqual(rsp.status_code, 403)

            # Make sure we're not logged in.
            self.assertNotIn('_auth_user_id', self.client.session)

    def test_post_assertion_replay_countermeasures(self) -> None:
        """Testing SAMLACSView POST replay attack countermeasures"""
        settings: JSONDict = {
            'saml_enabled': True,
        }

        with self.siteconfig_settings(settings):
            url = reverse('sso:saml:acs', kwargs={'backend_id': 'saml'})

            # First one should succeed and redirect us to the link user view.
            rsp = self.client.post(url, {})
            self.assertEqual(rsp.status_code, 302)

            # Second one should fail.
            rsp = self.client.post(url, {})
            self.assertEqual(rsp.status_code, 400)
            self.assertEqual(rsp.content,
                             b'SAML message IDs have already been used')


class SAMLSLSViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for SAMLSLSView."""

    def test_get_sls_replay_countermeasures(self) -> None:
        """Testing SAMLSLSView GET replay attack countermeasures"""
        class FakeAuth:
            def process_slo(*args, **kwargs):
                pass

            def get_last_message_id(self):
                return 'message-id'

            def get_last_request_id(self):
                return 'request-id'

            def get_last_error_reason(self):
                return None

        fake_auth = FakeAuth()

        self.spy_on(SAMLSLSView.get_saml_auth,
                    op=kgb.SpyOpReturn(fake_auth),
                    owner=SAMLSLSView)

        settings: JSONDict = {
            'saml_enabled': True,
        }

        with self.siteconfig_settings(settings):
            url = reverse('sso:saml:sls', kwargs={'backend_id': 'saml'})

            # First one should succeed and redirect us to the link user view.
            rsp = self.client.get(url, {})
            self.assertEqual(rsp.status_code, 302)

            # Second one should fail.
            rsp = self.client.get(url, {})
            self.assertEqual(rsp.status_code, 400)
            self.assertEqual(rsp.content,
                             b'SAML message IDs have already been used')



import base64
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from django.conf import settings


AES_BLOCK_SIZE = algorithms.AES.block_size // 8


def _create_cipher(iv, key):
    """Create a cipher for use in symmetric encryption/decryption.

    This will use AES encryption in CFB mode (using an 8-bit shift register)
    and a random IV.

    Args:
        iv (bytes):
            The random IV to use for the cipher.

        key (bytes):
            The encryption key to use.

    Returns:
        cryptography.hazmat.primitives.cipher.Cipher:
        The cipher to use for encryption/decryption.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    if not isinstance(key, bytes):
        raise TypeError('The encryption key must be of type "bytes", not "%s"'
                        % type(key))

    return Cipher(algorithms.AES(key),
                  modes.CFB8(iv),
                  default_backend())


def get_default_aes_encryption_key():
    """Return the default AES encryption key for the install.

    The default key is the first 16 characters (128 bits) of
    :django:setting:`SECRET_KEY`.

    Returns:
        bytes:
        The default encryption key.
    """
    return settings.SECRET_KEY[:16].encode('utf-8')


def aes_encrypt(data, key=None):
    """Encrypt data using AES encryption.

    This uses AES encryption in CFB mode (using an 8-bit shift register) and a
    random IV (which will be prepended to the encrypted value). The encrypted
    data will be decryptable using the :py:func:`aes_decrypt` function.

    Args:
        data (bytes):
            The data to encrypt. If a unicode string is passed in, it will be
            encoded to UTF-8 first.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        bytes:
        The resulting encrypted value, with the random IV prepended.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    if isinstance(data, str):
        data = data.encode('utf-8')

    iv = os.urandom(AES_BLOCK_SIZE)
    cipher = _create_cipher(iv, key or get_default_aes_encryption_key())
    encryptor = cipher.encryptor()

    return iv + encryptor.update(data) + encryptor.finalize()


def aes_decrypt(data, key=None):
    """Decrypt AES-encrypted data.

    This will decrypt an AES-encrypted value in CFB mode (using an 8-bit
    shift register). It expects the 16-byte cipher IV to be prepended to the
    string.

    This is intended as a counterpart for :py:func:`aes_encrypt`.

    Args:
        data (bytes):
            The data to decrypt.

        key (bytes, optional):
            The optional custom encryption key to use. This must match the key
            used for encryption. If not supplied, the default encryption key
            (from :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        bytes:
        The decrypted value.

    Raises:
        TypeError:
            One or more arguments had an invalid type.

        ValueError:
            The encryption key was not in the right format.
    """
    if not isinstance(data, bytes):
        raise TypeError('The data to decrypt must be of type "bytes", not "%s"'
                        % (type(data)))

    cipher = _create_cipher(data[:AES_BLOCK_SIZE],
                            key or get_default_aes_encryption_key())
    decryptor = cipher.decryptor()

    return decryptor.update(data[AES_BLOCK_SIZE:]) + decryptor.finalize()


def encrypt_password(password, key=None):
    """Encrypt a password and encode as Base64.

    The password will be encrypted using AES encryption in CFB mode (using an
    8-bit shift register), and serialized into Base64.

    Version Changed:
        4.0:
        The return type has been changed to :py:class:`unicode`, in order to
        improve the expected behavior on Python 3.

    Args:
        password (unicode or bytes):
            The password to encrypt. If a unicode string is passed in, it will
            be encoded to UTF-8 first.

        key (bytes, optional):
            The optional custom encryption key to use. If not supplied, the
            default encryption key (from
            :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        unicode:
        The encrypted password encoded in Base64.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    return base64.b64encode(aes_encrypt(password, key=key)).decode('utf-8')


def decrypt_password(encrypted_password, key=None):
    """Decrypt an encrypted password encoded in Base64.

    This will decrypt a Base64-encoded encrypted password (from
    :py:func:`encrypt_password`) into a usable password string.

    Version Changed:
        4.0:
        The return type has been changed to :py:class:`unicode`, in order to
        improve the expected behavior on Python 3.

    Args:
        encrypted_password (unicode or bytes):
            The Base64-encoded encrypted password to decrypt.

        key (bytes, optional):
            The optional custom encryption key to use. This must match the key
            used for encryption. If not supplied, the default encryption key
            (from :py:func:`get_default_aes_encryption_key)` will be used.

    Returns:
        unicode:
        The resulting password.

    Raises:
        ValueError:
            The encryption key was not in the right format.
    """
    result = (
        aes_decrypt(base64.b64decode(encrypted_password), key=key)
        .decode('utf-8')
    )

    if not result.isprintable():
        raise ValueError(
            'Unable to decrypt stored password: decrypted data contains '
            'unprintable characters. It is likely that the SECRET_KEY has '
            'changed.'
        )

    return result



"""Unit tests for WatchedResource.

Version Added:
    7.0.3
"""

from __future__ import annotations

import json
from typing import Any

from reviewboard.webapi.resources import resources
from reviewboard.webapi.tests.base import BaseWebAPITestCase
from reviewboard.webapi.tests.mimetypes import watched_mimetype
from reviewboard.webapi.tests.mixins import (
    BasicTestsMetaclass,
    BasicGetItemTestSetupState,
)
from reviewboard.webapi.tests.urls import get_watched_url


class ResourceTests(BaseWebAPITestCase, metaclass=BasicTestsMetaclass):
    """Unit tests for the WatchedResource.

    Version Added:
        7.0.3
    """

    basic_get_returns_json = False
    fixtures = ['test_users']
    sample_api_url = 'users/<username>/watched/'
    resource = resources.watched
    test_http_methods = ('GET',)

    def populate_get_item_test_objects(
        self,
        *,
        setup_state: BasicGetItemTestSetupState,
        **kwargs,
    ) -> None:
        """Populate objects for a GET item test.

        Args:
            setup_state (rbtools.webapi.tests.mixins.
                         BasicGetItemTestSetupState):
                The setup state for the test.

            **kwargs (dict):
                Additional keyword arguments for future expansion.
        """
        setup_state['url'] = get_watched_url(
            'doc', local_site_name=setup_state['local_site_name'])
        setup_state['mimetype'] = watched_mimetype
        setup_state['item'] = None

    def compare_item(
        self,
        item_rsp: bytes,
        obj: Any,
    ) -> None:
        """Compare an item's response payload to an object.

        Args:
            item_rsp (dict):
                The item payload from the response.

            obj (object):
                The object to compare to.
        """
        rsp = json.loads(item_rsp.decode())

        self.assertIn('links', rsp)
        self.assertIn('watched_review_groups', rsp['links'])
        self.assertIn('watched_review_requests', rsp['links'])



"""Root general comments API resource.

Version Added:
    5.0
"""

from django.contrib.auth.models import User
from django.db.models import Q
from djblets.util.decorators import augment_method_from
from djblets.webapi.decorators import webapi_request_fields
from djblets.webapi.fields import (BooleanFieldType,
                                   DateTimeFieldType,
                                   StringFieldType)

from reviewboard.reviews.models import GeneralComment
from reviewboard.webapi.resources.base_review_general_comment import \
    BaseReviewGeneralCommentResource


class RootGeneralCommentResource(BaseReviewGeneralCommentResource):
    """Provide information on general comments.

    This is a top level endpoint that allows you to list and query all
    general comments in the system, across different review requests.

    Version Changed:
        5.0.4:
        This was formally named ``general_comments`` on the root resource, but
        had to be renamed to ``all_general_comments`` in order to work around a
        conflict with URI templates.

    Version Added:
        5.0
    """

    added_in = '5.0'
    allowed_methods = ('GET',)
    model = GeneralComment
    link_name = 'all_general_comments'
    uri_template_name = 'all_general_comment'
    uri_object_key = None

    def get_queryset(self, request, is_list=False, *args, **kwargs):
        """Return a queryset for GeneralComment models.

        By default, this returns all general comments that are accessible
        to the requester.

        The queryset can be further filtered by one or more arguments
        in the URL. These are listed in :py:meth:`get_list`.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            is_list (bool, unused):
                Whether the coming HTTP request is request for list resources.

            *args (tuple):
                Additional positional arguments.

            **kwargs (dict):
                Additional keyword arguments.

        Returns:
            django.db.models.query.QuerySet:
            A queryset for GeneralComment models.
        """
        local_site = request.local_site
        q = Q()

        if 'is-reply' in request.GET:
            val = request.GET.get('is-reply')
            q &= Q(reply_to_id__isnull=(not val))

        if 'last-updated-from' in request.GET:
            q &= Q(timestamp__gte=request.GET.get('last-updated-from'))

        if 'last-updated-to' in request.GET:
            q &= Q(timestamp__lt=request.GET.get('last-updated-to'))

        if 'review-id' in request.GET:
            q &= Q(review=request.GET.get('review-id'))

        if 'review-request-id' in request.GET:
            review_request_id = request.GET.get('review-request-id')

            if local_site is None:
                q &= Q(review__review_request=review_request_id)
            else:
                q &= (Q(review__review_request__local_id=review_request_id) &
                      Q(review__review_request__local_site=local_site))

        if 'user' in request.GET:
            user = list((
                User.objects
                .filter(username=request.GET.get('user'))
                .values_list('pk', flat=True)
            ))

            if user:
                q &= Q(review__user=user[0])
            else:
                return self.model.objects.none()

        return self.model.objects.accessible(request.user,
                                             extra_query=q,
                                             local_site=local_site)

    @webapi_request_fields(
        optional={
            'is-reply': {
                'type': BooleanFieldType,
                'description': 'Determine whether to return general '
                               'comments that are replies or not.',
            },
            'last-updated-to': {
                'type': DateTimeFieldType,
                'description': "The date/time that all general "
                               "comments must be last updated before. This is "
                               "compared against the general "
                               "comment's ``timestamp`` field. This must be "
                               "a valid :term:`date/time format`.",
            },
            'last-updated-from': {
                'type': DateTimeFieldType,
                'description': "The earliest date/time the general "
                               "comments could be last updated. This is "
                               "compared against the general "
                               "comment's ``timestamp`` field. This must "
                               "be a valid :term:`date/time format`.",
            },
            'review-id': {
                'type': StringFieldType,
                'description': 'The ID of the review that the general '
                               'comments must belong to.',
            },
            'review-request-id': {
                'type': StringFieldType,
                'description': 'The ID of the review request that the general '
                               'comments must belong to.',
            },
            'user': {
                'type': StringFieldType,
                'description': 'The username of the user that created the '
                               'general comments.',
            },
        },
        allow_unknown=True
    )
    @augment_method_from(BaseReviewGeneralCommentResource)
    def get_list(self, *args, **kwargs):
        """Return the list of general comments."""
        pass


root_general_comment_resource = RootGeneralCommentResource()



"""Set null=False for depends_on fields.

This only affects the stored signature, and should not impact the database
schema.

Version Added:
    4.0
"""

from django_evolution.mutations import ChangeField


MUTATIONS = [
    ChangeField('ReviewRequest', 'depends_on', initial=None, null=False),
    ChangeField('ReviewRequestDraft', 'depends_on', initial=None, null=False),
]



"""DiffCommit model definition."""

from __future__ import annotations

from typing import ClassVar, Optional

from dateutil.tz import tzoffset
from django.db import models
from django.utils import timezone
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _
from djblets.db.fields import JSONField
from typing_extensions import TypedDict

from reviewboard.diffviewer.diffutils import get_total_line_counts
from reviewboard.diffviewer.managers import DiffCommitManager
from reviewboard.diffviewer.models.diffset import DiffSet
from reviewboard.diffviewer.validators import (COMMIT_ID_LENGTH,
                                               validate_commit_id)


class SerializedDiffCommit(TypedDict):
    """Serialized version of a DiffCommit.

    Version Added:
        7.0
    """

    #: The name of the author of the commit.
    author_name: str

    #: The SCM-specific commit ID of the commit.
    commit_id: str

    #: The commit message.
    commit_message: str

    #: The PK of the :py:class:`DiffCommit` object.
    id: int

    #: The SCM-specific ID of the parent of the commit.
    parent_id: str


class DiffCommit(models.Model):
    """A representation of a commit from a version control system.

    A DiffSet on a Review Request that represents a commit history will have
    one or more DiffCommits. Each DiffCommit will have one or more associated
    FileDiffs (which also belong to the parent DiffSet).

    The information stored herein is intended to fully represent the state of
    a single commit in that history. The series of DiffCommits can be used to
    re-create the original series of commits posted for review.
    """

    #: The maximum length of the author_name and committer_name fields.
    NAME_MAX_LENGTH = 256

    #: The maximum length of the author_email and committer_email fields.
    EMAIL_MAX_LENGTH = 256

    #: The date format that this model uses.
    ISO_DATE_FORMAT = '%Y-%m-%d %H:%M:%S%z'

    filename = models.CharField(
        _('File Name'),
        max_length=256,
        help_text=_('The original file name of the diff.'))

    diffset = models.ForeignKey(DiffSet,
                                on_delete=models.CASCADE,
                                related_name='commits')

    author_name = models.CharField(
        _('Author Name'),
        max_length=NAME_MAX_LENGTH,
        help_text=_('The name of the commit author.'))
    author_email = models.CharField(
        _('Author Email'),
        max_length=EMAIL_MAX_LENGTH,
        help_text=_('The e-mail address of the commit author.'))
    author_date_utc = models.DateTimeField(
        _('Author Date'),
        help_text=_('The date the commit was authored in UTC.'))
    author_date_offset = models.IntegerField(
        _('Author Date UTC Offset'),
        help_text=_("The author's UTC offset."))

    committer_name = models.CharField(
        _('Committer Name'),
        max_length=NAME_MAX_LENGTH,
        help_text=_('The name of the committer (if applicable).'),
        null=True,
        blank=True)
    committer_email = models.CharField(
        _('Committer Email'),
        max_length=EMAIL_MAX_LENGTH,
        help_text=_('The e-mail address of the committer (if applicable).'),
        null=True,
        blank=True)
    committer_date_utc = models.DateTimeField(
        _('Committer Date'),
        help_text=_('The date the commit was committed in UTC '
                    '(if applicable).'),
        null=True,
        blank=True)
    committer_date_offset = models.IntegerField(
        _('Committer Date UTC Offset'),
        help_text=_("The committer's UTC offset (if applicable)."),
        null=True,
        blank=True)

    commit_message = models.TextField(
        _('Description'),
        help_text=_('The commit message.'))

    commit_id = models.CharField(
        _('Commit ID'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id],
        help_text=_('The unique identifier of the commit.'))

    parent_id = models.CharField(
        _('Parent ID'),
        max_length=COMMIT_ID_LENGTH,
        validators=[validate_commit_id],
        help_text=_('The unique identifier of the parent commit.'))

    #: A timestamp used for generating HTTP caching headers.
    last_modified = models.DateTimeField(
        _('Last Modified'),
        default=timezone.now)

    extra_data = JSONField(null=True)

    objects: ClassVar[DiffCommitManager] = DiffCommitManager()

    @property
    def author(self):
        """The author's name and e-mail address.

        This is formatted as :samp:`{author_name} <{author_email}>`.
        """
        return self._format_user(self.author_name, self.author_email)

    @property
    def author_date(self):
        """The author date in its original timezone."""
        tz = tzoffset(None, self.author_date_offset)
        return self.author_date_utc.astimezone(tz)

    @author_date.setter
    def author_date(self, value):
        """Set the author date.

        Args:
            value (datetime.datetime):
                The date to set.
        """
        self.author_date_utc = value

        if value is not None:
            self.author_date_offset = value.utcoffset().total_seconds()
        else:
            self.author_date_offset = None

    @property
    def committer(self):
        """The committer's name and e-mail address (if applicable).

        This will be formatted as :samp:`{committer_name} <{committer_email}>`
        if both :py:attr:`committer_name` and :py:attr:`committer_email` are
        set. Otherwise, it be whichever is defined. If neither are defined,
        this will be ``None``.
        """
        return self._format_user(self.committer_name, self.committer_email)

    @property
    def committer_date(self):
        """The committer date in its original timezone.

        If the commit has no committer, this will be ``None``.
        """
        if self.committer_date_offset is None:
            return None

        tz = tzoffset(None, self.committer_date_offset)
        return self.committer_date_utc.astimezone(tz)

    @committer_date.setter
    def committer_date(self, value):
        """Set the committer date.

        Args:
            value (datetime.datetime):
                The date to set.
        """
        self.committer_date_utc = value

        if value is not None:
            self.committer_date_offset = value.utcoffset().total_seconds()
        else:
            self.committer_date_offset = None

    @cached_property
    def summary(self):
        """The first line of the commit message."""
        summary = self.commit_message

        if summary:
            summary = summary.split('\n', 1)[0].strip()

        return summary

    @cached_property
    def commit_message_body(self) -> Optional[str]:
        """The body of a commit message.

        This will contain the content following a commit summary, if any.

        Version Added:
            6.0

        Type:
            str
        """
        body: Optional[str] = None
        message = self.commit_message

        if message:
            parts = message.split('\n', 1)

            if len(parts) > 1:
                body = parts[1].strip() or None

        return body

    @cached_property
    def summary_truncated(self):
        """The first line of the commit message, truncated to 80 characters."""
        summary = self.summary

        if len(summary) > 80:
            summary = summary[:77] + '...'

        return summary

    def serialize(self) -> SerializedDiffCommit:
        """Serialize to a dictionary.

        Returns:
            dict:
            A dictionary representing this commit.
        """
        return {
            'author_name': self.author_name,
            'commit_id': self.commit_id,
            'commit_message': self.commit_message,
            'id': self.pk,
            'parent_id': self.parent_id,
        }

    def get_total_line_counts(self):
        """Return the total line counts of all child FileDiffs.

        Returns:
            dict:
            A dictionary with the following keys:

            * ``raw_insert_count``
            * ``raw_delete_count``
            * ``insert_count``
            * ``delete_count``
            * ``replace_count``
            * ``equal_count``
            * ``total_line_count``

            Each entry maps to the sum of that line count type for all child
            :py:class:`FileDiffs
            <reviewboard.diffviewer.models.filediff.FileDiff>`.
        """
        return get_total_line_counts(self.files.all())

    def __str__(self):
        """Return a human-readable representation of the commit.

        Returns:
            unicode:
            The commit ID and its summary (if available).
        """
        if self.summary:
            return '%s: %s' % (self.commit_id, self.summary)

        return self.commit_id

    def _format_user(self, name, email):
        """Format a name and e-mail address.

        Args:
            name (unicode):
                The user's name.

            email (unicode):
                The user's e-mail address.

        Returns:
            unicode:
            A pretty representation of the user and e-mail, or ``None`` if
            neither are defined.
        """
        if name and email:
            return '%s <%s>' % (name, email)
        elif name:
            return name
        elif email:
            return email

        return None

    class Meta:
        app_label = 'diffviewer'
        db_table = 'diffviewer_diffcommit'
        verbose_name = _('Diff Commit')
        verbose_name_plural = _('Diff Commits')
        unique_together = ('diffset', 'commit_id')
        ordering = ('pk',)



"""Added DiffCommit-related fields to DiffSet and FileDiff.

Version Added:
    4.0
"""

from django_evolution.mutations import AddField
from django.db import models
from djblets.db.fields import RelationCounterField


MUTATIONS = [
    AddField('DiffSet', 'commit_count', RelationCounterField, null=True),
    AddField('DiffSet', 'file_count', RelationCounterField, null=True),
    AddField('DiffCommit', 'file_count', RelationCounterField, null=True),
    AddField('FileDiff', 'commit', models.ForeignKey, null=True,
             related_model='diffviewer.DiffCommit'),
]
